
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */

#include "gm_bsd_ether.h"

/*
 * For more information about FreeBSD network drivers, look at a few
 * of the better drivers in the FreeBSD tree.  In particular,
 * sys/dev/bge/* and sys/dev/fxp/* are good examples.  Questions
 * regarding freebsd netoworking are best addressed to
 * the freebsd-net@FreeBSD.ORG mailing list.
 */

struct gm_bsd_ether_sc *gm_bsd_ether_sc[GM_ARCH_MAX_INSTANCE] = { 0 };
int gm_ether_count = 0;
static int gm_mbufs_loaned = 0;
static int gm_mbuf_wakeup = 0;

struct mbuf *bogus_m = NULL;
struct gm_bsd_ether_arch_dma_info bogus_dma;


/* 
 *  allocates a new "softc" struct for the gm ether device
 */

static struct gm_bsd_ether_sc *
gm_freebsd_ether_alloc_sc (int unit)
{
  struct gm_bsd_ether_sc *sc;

  if (gm_bsd_ether_sc[unit] != NULL)
    {
      GM_WARN (("attempt to attach myri%d twice failed\n", unit));
      return NULL;
    }

  sc = gm_malloc(sizeof (*sc));

  bzero(sc, sizeof (struct gm_bsd_ether_sc));
  if (sc != NULL)
    {
      gm_bsd_ether_sc[unit] = sc;
    }
  gm_ether_count++;
  return sc;
}

/* 
 * free a softc
 */

static void
gm_freebsd_ether_dealloc_sc (int unit)
{
  struct gm_bsd_ether_sc *sc;

  sc = gm_bsd_ether_sc[unit];
  gm_bsd_ether_sc[unit] = NULL;
  gm_free (sc);
}

/*
 * called by ether_ioctl to "initialize" an interface.  We must
 * supply the function pointer, but we do not use this feature.
 */

static void
gm_freebsd_ether_init (void *sc_ptr)
{
}

#if (__FreeBSD_version >= 500000 )
/*
 * This is the transmit routine.  Obtain the sleep lock
 * for the driver and call the common transmit routine.
 */
static void
gm_freebsd_ether_start (struct ifnet *ifp)
{
  struct gm_ether_softc *sc = ifp->if_softc;

  gm_bsd_ether_arch_sc_lock (ifp->if_softc);
  gm_bsd_ether_start (ifp);
  gm_bsd_ether_arch_sc_unlock (ifp->if_softc);
}

#endif

/*
 * creates a new ethernet driver, initialzes it, and attaches it to the
 * network stack
 */
int
gm_freebsd_ether_attach (gm_instance_state_t * is)
{
  int i, unit;
  struct ifnet *ifp;
  struct gm_bsd_ether_sc *sc;

  if (GM_MAX_ETHERNET_GATHER_CNT * MCLBYTES <= GM_ETHERNET_MTU)
    {
      GM_WARN (("GM_MAX_ETHERNET_GATHER_CNT (%d) is too small to hold a\n"
		"%d byte long mbuf chain consisting of %d byte mbufs\n"
		"Each transmit of %d bytes or more will have to be copied\n"
		"Consider increasing MCLSHIFT and rebuilding your kernel & GM\n"
		"or rebuilding GM using a 32-bit gm_dp_t\n",
		GM_MAX_ETHERNET_GATHER_CNT, GM_ETHERNET_MTU, MCLBYTES,
		(GM_MAX_ETHERNET_GATHER_CNT - 1) * MCLBYTES));
    }
  if (MCLBYTES != mbstat.m_mclbytes)
    {
      GM_WARN (("module MCLBYTES (%d) != kernel MCLBYTES (%d)\n", MCLBYTES,
		mbstat.m_mclbytes));
      return EINVAL;
    }
  unit = device_get_unit (is->arch.dev);
  sc = gm_freebsd_ether_alloc_sc (unit);
  if (sc == NULL)
    {
      return ENOMEM;
    }
  sc->is = is;
  ifp = &sc->is_if;
  ifp->if_softc = sc;
  ifp->if_addrlen = 6;
  for (i = 0; i < 6; i++)
    {
      sc->is_addr[i] = is->lanai.eeprom.lanai_board_id[i];
    }

  ifp->if_name = "myri";
  ifp->if_unit = unit;

  /* leave extra space to prepend myritype */
  ifp->if_hdrlen = sizeof (struct ether_header) + 8;
  ifp->if_mtu = GM_IP_MTU;
  ifp->if_timer = 0;
  ifp->if_ioctl = gm_freebsd_ether_ioctl;
  ifp->if_watchdog = NULL;
  ifp->if_init = gm_freebsd_ether_init;
#if (__FreeBSD_version < 500000)
  ifp->if_start = gm_bsd_ether_start;
#else
  mtx_init (&sc->arch.mutex, "myri", NULL, MTX_DEF);
  ifp->if_start = gm_freebsd_ether_start;
#endif
  ifp->if_output = ether_output;
  ifp->if_flags = IFF_BROADCAST | IFF_MULTICAST | IFF_SIMPLEX;
  ifp->if_baudrate = 1000000000;
  ifp->if_snd.ifq_maxlen = GM_SEND_RING_CNT;
#if (__FreeBSD_version >= 410000)
  ether_ifattach (ifp, ETHER_BPF_SUPPORTED);
#else
  if_attach (ifp);
  ether_ifattach (ifp);
#endif
  /*ether_ifattach resets mtu */
  ifp->if_mtu = GM_IP_MTU;
  return 0;

}

static int
gm_freebsd_ether_ioctl (struct ifnet *ifp, unsigned long cmd, caddr_t data)
{
  struct gm_bsd_ether_sc *sc = ifp->if_softc;
  struct ifreq *ifr = (struct ifreq *) data;
  int s, error;

  s = splimp ();

  switch (cmd)
    {

    case SIOCSIFADDR:
    case SIOCGIFADDR:
      error = ether_ioctl (ifp, cmd, data);
      break;

    case SIOCSIFMTU:
      if (ifr->ifr_mtu > GM_IP_MTU)
	{
	  error = EINVAL;
	}
      else
	{
	  ifp->if_mtu = ifr->ifr_mtu;
	  error = 0;
	}
      break;
    case SIOCSIFFLAGS:
      if (ifp->if_flags & IFF_UP)
	{
	  error = gm_bsd_ether_open (sc);
	  ifp->if_flags |= IFF_RUNNING;
	}
      else
	{
	  ifp->if_flags &= ~IFF_RUNNING;
	  error = gm_bsd_ether_close (sc);
	}
      break;
    case SIOCADDMULTI:
      /* fall through */
    case SIOCDELMULTI:
      error = 0;
      break;
    default:
      error = ENOTTY;
    }

  splx (s);
  return error;
}

MALLOC_DEFINE (M_GMBUF, "gm_mbuf", "gm_mbuf -- buffers for ethernet receive");

#if GM_PREVENT_NATIVE_MALLOC_USE
#undef malloc
#undef free
#endif /* GM_PREVENT_MALLOC_USE */


static void
gm_freebsd_ether_rxfree (
#if (__FreeBSD_version < 500000)
			 caddr_t ptr, u_int ingnored
#else
			 void *ptr, void *args
#endif
			 )
{
  unsigned long *ref_count;

  ref_count = (unsigned long *) ptr;
  if (*ref_count == 0)
    {
      FREE (ptr, M_GMBUF);
      gm_mbufs_loaned--;
      if (gm_mbuf_wakeup)
	wakeup (&gm_mbufs_loaned);
    }
  else
    {
      *ref_count -= 1;
    }
}

#if (__FreeBSD_version < 500000)
static void
gm_freebsd_ether_rxref (caddr_t ptr, u_int ingnored)
{
  unsigned long *ref_count;

  ref_count = (unsigned long *) ptr;
  *ref_count += 1;
}

#endif /* __FreeBSD_version < 500000 */


int fail_test = 0;

struct mbuf *
gm_bsd_ether_arch_rxalloc (int how, int max_segs)
{
  struct mbuf *m_head, *mp, *m_tail;
  int len;
  unsigned long *ref_count;
  char *buf;

  len = GM_ETHERNET_MTU;

#if (__FreeBSD_version < 470000)
  /* get the first header mbuf */
  MGETHDR (mp, how, MT_DATA);
  if (mp == NULL)
    {
      return NULL;
    }

  /* make it into a 2K cluster mbuf */
  MCLGET (mp, how);
  if ((mp->m_flags & M_EXT) == 0)
    {
      m_free (mp);
      return NULL;
    }
#else /* use new helper function */
  mp = m_getcl(how, MT_DATA, M_PKTHDR);
  if (mp == NULL)
    {
      return NULL;
    }
#endif
  /* start building the chain of PAGE_SIZE mbufs */
  m_head = m_tail = mp;
  mp->m_len = MCLBYTES;
  len -= MCLBYTES;

  while (len > 0)
    {
      if (MCLBYTES * max_segs >= GM_ETHERNET_MTU)
	{
#if (__FreeBSD_version < 470000)
	  /* get the next header mbuf */
	  MGET (mp, how, MT_DATA);
	  if (mp == NULL)
	    {
	      goto abort_with_cluster;
	    }
	  /* attach to chain */
	  m_tail->m_next = mp;
	  m_tail = mp;
	  /* use a 2K "cluster" */
	  MCLGET (mp, how);
	  if ((mp->m_flags & M_EXT) == 0)
	    {
	      goto abort_with_cluster;
	    }
#else
	  mp = m_getcl(how, MT_DATA, 0);
	  if (mp == NULL)
	    {
	      goto abort_with_cluster;
	    }
	  m_tail->m_next = mp;
	  m_tail = mp;
#endif
	  mp->m_len = MCLBYTES;
	  len -= MCLBYTES;
	}
      else
	{
	  /* get the next header mbuf */
	  MGET (mp, how, MT_DATA);
	  if (mp == NULL)
	    {
	      goto abort_with_cluster;
	    }
	  /* attach to chain */
	  m_tail->m_next = mp;
	  m_tail = mp;

	  /* make it into a PAGE_SIZEed external mbuf */
	  MALLOC (buf, caddr_t, PAGE_SIZE, M_GMBUF, how);
	  if (mp->m_ext.ext_buf == NULL)
	    {
	      goto abort_with_cluster;
	    }
#if (__FreeBSD_version < 500000)
	  mp->m_ext.ext_buf = buf;
	  mp->m_ext.ext_free = gm_freebsd_ether_rxfree;
	  mp->m_ext.ext_ref = gm_freebsd_ether_rxref;
	  mp->m_flags |= M_EXT;
	  mp->m_ext.ext_size = PAGE_SIZE - sizeof (long);
#else
	  MEXTADD(mp, buf, PAGE_SIZE -sizeof (long), 
		  gm_freebsd_ether_rxfree, buf, 0, EXT_NET_DRV);
#endif
	  mp->m_len = PAGE_SIZE - sizeof (long);
	  mp->m_data = mp->m_ext.ext_buf + sizeof (long);
	  len -= mp->m_len;
	  ref_count = (unsigned long *) mp->m_ext.ext_buf;
	  *ref_count = 0;
	  gm_mbufs_loaned++;
	  /* this routine is built on the assumption that the
	   * buffers FreeBSD's malloc will give us are not going to
	   * cross page boundaries.  Verify this with an assert
	   */
	  gm_always_assert (trunc_page ((vm_offset_t) mp->m_ext.ext_buf) ==
			    (vm_offset_t) mp->m_ext.ext_buf);
	}
    }
  return m_head;

 abort_with_cluster:
  m_freem (m_head);
  return (NULL);
}

#if GM_PREVENT_NATIVE_MALLOC_USE
#define malloc GM_DO_NOT_USE (malloc)
#define free GM_DO_NOT_USE (free)
#endif /* GM_PREVENT_MALLOC_USE */

/* 
 * Encapsulate an mbuf chain into an ethernet segment descriptor.
 * Since FreeBSD doesn't have real DMA mapping support, this is easy.
 * This routine returns the number of segments in the ethernet segment
 * descriptor.
 */

int
gm_bsd_ether_arch_encap (struct gm_bsd_ether_sc *sc, struct mbuf *m_head,
		struct gm_bsd_ether_arch_dma_info *dma)
{
  int count;
  struct mbuf *m;

  count = 0;
  for (m = m_head; m != NULL; m = m->m_next)
    {
      if (m->m_len == 0)
	continue;

      dma->desc[count].ptr = gm_hton_dp (vtophys (mtod (m, caddr_t)));
      dma->desc[count].len = gm_hton_u32 (m->m_len);
      count++;
    }
  dma->nsegs = count;
  return count;
}

void
gm_freebsd_ether_detach (gm_instance_state_t * is)
{
  struct gm_bsd_ether_sc *sc;
  struct ifnet *ifp, *gm_ifp;
  int s, ifp_valid;

  sc = gm_bsd_ether_sc[device_get_unit (is->arch.dev)];
  /* if we failed to attach, then there's nothing to detach */
  if (sc == NULL)
    {
      return;
    }

  s = splnet ();

  /* search for our ifnet in the global list */
  ifp_valid = 0;
  gm_ifp = &sc->is_if;
  TAILQ_FOREACH (ifp, &ifnet, if_link)
    {
      if (ifp == gm_ifp)
	{
	  ifp_valid = 1;
	  break;
	}
    }

  if (ifp_valid)
    {
#if (__FreeBSD_version >= 410000)
      ether_ifdetach (gm_ifp, ETHER_BPF_SUPPORTED);
#else /* no ether_ifdetach() available, so do it by hand */
      struct ifaddr *ia = 0;
      struct in_ifaddr *in_ia = 0;

      TAILQ_FOREACH (ia, &ifp->if_addrhead, ifa_link)
	{
	  if (ia->ifa_addr->sa_family != AF_INET)
	    continue;
	  in_ia = (struct in_ifaddr *) ia;
	  in_ifscrub (gm_ifp, in_ia);
	  TAILQ_REMOVE (&in_ifaddrhead, in_ia, ia_link);
	}
      if_detach (gm_ifp);
#endif /* (__FreeBSD_version >= 410000) */
    }

  splx (s);

  /* 
   * the system no longer knows about the interface, close it &
   * deallocate it 
   */

  if (sc->open)
    {
      gm_bsd_ether_close (sc);
    }

  while (gm_mbufs_loaned != 0)
    {
      gm_mbuf_wakeup = 1;
      GM_NOTE (("%d mbufs still remain to be freed, sleeping\n",
		gm_mbufs_loaned));
      tsleep (&gm_mbufs_loaned, PUSER, "gm_freebsd_ether_detach", 10 * hz);
    }

#if (__FreeBSD_version >= 500000)  
  mtx_destroy (&sc->arch.mutex);
#endif
  gm_freebsd_ether_dealloc_sc (device_get_unit (is->arch.dev));
}




/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
