
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */

/*
 * For more information regarding mbufs, the ifnet structure and BSD
 * networking in general, _The Design and Implementation of the 4.4BSD
 * Operating System_ by McKusic, Bostic, Karels & Quarterman, 
 * (Addison Wesley, 1996) is recommended reading, as is the network
 * code of the open source BSDs (in sys/net and sys/netinet).
 */

#include "gm_bsd_ether.h"
int gm_bsd_ether_xmit_copies = 0;

/* 
 * debugging aid to print all the valid data in an mbuf chain
 */
void
gm_bsd_ether_print_mbuf (struct mbuf *m)
{
  int i, np, length, tot_len;
  struct mbuf *t;

  tot_len = 0;
  length = m->m_pkthdr.len;
  printf ("mbuf length is %d\n", length);

  np = 0;

  for (t = m; t; t = t->m_next)
    {
      printf ("piece %d, len %d\n", np++, t->m_len);
      if (t->m_len == 0)
	continue;


      for (i = 0; (i < 20 /*t->m_len */ ) && (tot_len < length);
	   i++, tot_len++)
	printf ("%02x%c", 0xff & ((char *) (t->m_data))[i],
		i % 8 == 7 ? '\n' : ' ');
      printf ("\n");
    }

}


/*
 * transmit a packet. must be called with all required locks held & at
 * splimp
 */

void
gm_bsd_ether_start (struct ifnet *ifp)
{
  struct gm_bsd_ether_sc *sc;
  struct mbuf *m_head, *m;
  struct gm_bsd_ether_pkt *pkt;
  unsigned char *address;
  int broadcast, index, must_copy, nsegs, prepend;

#if (GM_OS_OSF1 == 1)
  sc = gm_bsd_ether_sc[ifp->if_unit];
#else
  sc = ifp->if_softc;
#endif

  if ((!sc->open) || (!sc->is->lanai.running))
    return;

  while (sc->sreq - sc->sdone < GM_SEND_RING_CNT)
    {
      index = sc->sreq & GM_SEND_RING_MAX_INDEX;
      pkt = sc->send_ring + index;

      IF_DEQUEUE (&ifp->if_snd, m_head);
      if (m_head == NULL)
	return;

      must_copy = 0;
      nsegs = 0;
      for (m = m_head; m != NULL; m = m->m_next)
	{
	  if (m->m_len == 0)
	    continue;
	  nsegs++;
	  if (nsegs > GM_MAX_ETHERNET_GATHER_CNT)
	    must_copy++;
#if (GM_EMULATE_BYTE_DMAS == 0)
	  if (!GM_DMA_ALIGNED (mtod (m, void *)))
	    must_copy++;
#endif
	}
      if (must_copy)
	{
	  gm_bsd_ether_xmit_copies++;
	  m = gm_m_dup (m_head);
	  if (m == NULL)
	    {
	      goto drop;
	    }
	  m_freem (m_head);
	  m_head = m;
	}
      address = mtod (m_head, unsigned char *);

      broadcast = address[0] & 0x1;
      gm_bsd_ether_arch_bpf_xmit (ifp, m_head);


      /* if we have enough space, prepend the myritype to save
       *  some cycles in the firmware
       */
      if (GM_EMULATE_BYTE_DMAS && M_LEADINGSPACE (m_head) >= sizeof (gm_u16_t))
	{
	  prepend = 0;
	  M_PREPEND (m_head, sizeof (gm_u16_t), M_NOWAIT);
	  *(gm_u16_t *) mtod (m_head, caddr_t) =
	    __gm_hton_u16 (GM_ETHERNET_PACKET_TYPE);
	}
      else
	{
	  prepend = 1;
	}

      if (!gm_bsd_ether_arch_encap (sc, m_head, &pkt->dma))
	{
	  goto drop;
	}

      pkt->m = m_head;

      /* update the send_ring prior to actually sending so the 
       * interrupt handler is sure to have up-to-date send_ring
       * info 
       */
      sc->sreq += 1;
      if (broadcast)
	{
	  gm_ethernet_broadcast (sc->port, pkt->dma.nsegs, pkt->dma.desc,
				 prepend);
	}
      else
	{
	  gm_ethernet_send (sc->port, pkt->dma.nsegs, pkt->dma.desc, address,
			    prepend);
	}

    }

  /* if we got here, then we filled up the send ring, so set the
   * OACTIVE flag to keep the stack from calling us again. 
   * The transmit complete interrupt handler will clear this flag
   * and call us as soon as space becomes available in the send ring
   */
  ifp->if_flags |= IFF_OACTIVE;

  return;

 drop:

  IF_DROP (&ifp->if_snd);
  m_freem (m_head);

}


/* transmit complete interrupt handler, driver is already at splimp
 * via main interrupt handler, so we just need to grab the lock */

void
gm_bsd_ether_txeof (void *ptr)
{
  struct gm_bsd_ether_sc *sc = (struct gm_bsd_ether_sc *) ptr;
  struct ifnet *ifp = &sc->is_if;
  struct gm_bsd_ether_pkt *pkt;
  int index;

  gm_bsd_ether_arch_sc_lock (&sc->arch);

  if (sc->sreq - sc->sdone == 0)
    {
      GM_WARN (("myri%d: Internal error: SENT_EVENT but no mbuf "
		"in send_ring!??,sreq=%u,sdone=%u\n", ifp->if_unit, sc->sreq,
		sc->sdone));
      gm_bsd_ether_arch_sc_unlock (&sc->arch);
      return;
    }
  index = sc->sdone & GM_SEND_RING_MAX_INDEX;
  pkt = sc->send_ring + index;
  ifp->if_opackets++;
  gm_bsd_ether_arch_incr_tx_counters (&sc->arch, pkt->m);

  gm_bsd_ether_arch_unencap (sc, &pkt->dma);
  m_freem (pkt->m);
  pkt->m = NULL;
  sc->sdone += 1;

  /* restart xmit if it is waiting for space in the send ring */
  if (ifp->if_flags & IFF_OACTIVE)
    {
      ifp->if_flags &= ~IFF_OACTIVE;
      gm_bsd_ether_start (ifp);
    }
  gm_bsd_ether_arch_sc_unlock (&sc->arch);
}

void
gm_bsd_ether_rxeof (void *ptr, unsigned int ulen, gm_u16_t csum)
{
  struct gm_bsd_ether_sc *sc = (struct gm_bsd_ether_sc *) ptr;
  struct ifnet *ifp = &sc->is_if;
  struct mbuf *m, *m_head, *m_new;
  struct gm_bsd_ether_pkt *pkt;
  struct ether_header *eh;
  int index, len;
  gm_u16_t myritype;

  gm_bsd_ether_arch_sc_lock (&sc->arch);
  index = sc->rdone & GM_RECV_RING_MAX_INDEX;
  sc->rdone += 1;
  pkt = sc->recv_ring + index;
  m_head = pkt->m;

  if (m_head == bogus_m)
    {
      /* 
       * used the "bogus" mbuf chain previous, data may be bad. Hope
       * we have enough resources & try to alloc a new chain this time
       */
      goto abort_need_realloc;
    }
  len = ulen;
  if (len < (sizeof (struct ether_header) + 2))
    {
      GM_WARN (("myri%d: received invalid packet: len=%d\n", 
		ifp->if_unit, len));
      goto abort_need_restock;
    }
  /* check the type, done here to keep m_head in a register */
  myritype = *(gm_u16_t *) mtod (m_head, caddr_t);
  if (myritype != __gm_hton_u16 (GM_ETHERNET_PACKET_TYPE))
    {
      GM_WARN (("myri%d: received packet with strange header:0x%x\n",
		ifp->if_unit, myritype));
      goto abort_need_restock;
    }
  if (GM_OS_FREEBSD == 0 && len < MHLEN)
    {
      /* copy packet rather than waste space, this also avoids
       * calling dma_map_{un,}load again & improves small packet
       * latency */
      MGETHDR (m_new, M_NOWAIT, MT_DATA);
      if (m_new == NULL)
	{
	  goto abort_need_restock;
	}
      else
	{
	  bcopy (mtod (m_head, char *), mtod (m_new, char *), len);

	  m_new->m_len = len;
	  m_head = m_new;
	  /* recycle m_head (pkt->m) */
	  _gm_ethernet_provide_scatter_list (sc->port, pkt->dma.nsegs,
					     pkt->dma.desc);
	}
    }
  else 	/* don't copy, but pass the data up the stack */
    {  

      /* alloc new rx buffer */
      m_new = gm_bsd_ether_arch_rxalloc (M_NOWAIT, GM_MAX_ETHERNET_SCATTER_CNT);
      if (m_new == NULL)
	{
	  goto abort_need_restock;
	}
      gm_bsd_ether_arch_unencap (sc, &pkt->dma);
      if (gm_bsd_ether_arch_encap (sc, m_new, &pkt->dma))
	{
	  pkt->m = m_new;
	  _gm_ethernet_provide_scatter_list (sc->port, pkt->dma.nsegs,
					     pkt->dma.desc);
	}
      else
	{
	  /* if the encapsulation fails, we have already unloaded the
	   * old DMA map, so we have no choice but to attach a known
	   * "bogus" mbuf chain so the lanai has buffers to dma into
	   */
	  m_freem (m_new);
	  pkt->m = bogus_m;
	  _gm_ethernet_provide_scatter_list (sc->port, bogus_dma.nsegs,
					     bogus_dma.desc);
	}
#if (GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED == 1)
      /* grab the HW checksum to save the IP stack from doing
       * software checksums. Note that we do this only when
       * we're not already touching the data by doing a copy 
       */
      gm_mbuf_rx_checksum (m_head, ulen, csum);
#endif
    }

  /* find the ether header */
  eh = (struct ether_header *) (mtod (m_head, char *) + 2);

#if (GM_OS_OSF1 == 1)
  /* tru64's ether_input doesn't swizzle the ether_type */
  eh->ether_type = htons (eh->ether_type);
#endif
  /* point the mbuf data at the ethernet payload & adjust len */
  ((char *)(m_head->m_data)) += sizeof (struct ether_header) + 2;
  m_head->m_len -= (sizeof (struct ether_header) + 2);
  len -= (sizeof (struct ether_header) + 2);


  /* let the upper layers know the len of the packet */
  m_head->m_pkthdr.len = len;
  m_head->m_pkthdr.rcvif = ifp;

  /* chop off excess elements of mbuf chain */
  for (m = m_head; (len > 0) && (m != NULL); m = m->m_next)
    {
      if (m->m_len >= len)
	{
	  /* adjust length & free rest of chain */
	  m->m_len = len;
	  if (m->m_next != NULL)
	    {
	      m_freem (m->m_next);
	      m->m_next = NULL;
	    }
	}
      len -= m->m_len;
    }

  /* update the interface counters */
  ifp->if_ipackets++;
  gm_bsd_ether_arch_incr_rx_counters (&sc->arch, ulen, eh);
  /* pass the packet up to the upper layers */
#if (GM_OS_MACOSX == 1)
  if (GM_MACOSX_VERSION == 101)
  {
    m_head->m_pkthdr.aux = (void *)ifp;
  }
  m_head->m_pkthdr.header = eh;
  gm_bsd_ether_arch_bpf_rx(ifp, m_head);
  dlil_input(ifp, m_head, NULL);
#else
  ether_input (ifp, eh, m_head);
#endif
  gm_bsd_ether_arch_sc_unlock (&sc->arch);
  return;			/* done with this packet, return */


  /* error handling code, expects old mbuf in m_head & valid dma
   * for m_head in pkt->dma. */

 abort_need_realloc:		/* get here if bogus_m was used previously */
  m_head = gm_bsd_ether_arch_rxalloc (M_NOWAIT, GM_MAX_ETHERNET_SCATTER_CNT);
  if (m_head == NULL)
    {
      m_head = bogus_m;
    }
  else
    {
      if (!gm_bsd_ether_arch_encap (sc, m_head, &pkt->dma))
	{
	  m_freem (m_head);
	  m_head = bogus_m;
	}
    }
 abort_need_restock:
  /* restock the lanai, re-using mbuf chain in m_head
     or the bogus dma pages in bogus_dma */
  if (m_head == bogus_m)
    {
      _gm_ethernet_provide_scatter_list (sc->port, bogus_dma.nsegs,
					 bogus_dma.desc);
    }
  else
    {
      _gm_ethernet_provide_scatter_list (sc->port, pkt->dma.nsegs,
					 pkt->dma.desc);
    }

  ifp->if_ierrors++;
  gm_bsd_ether_arch_sc_unlock (&sc->arch);


}


int
gm_bsd_ether_open (struct gm_bsd_ether_sc *sc)
{
  struct mbuf *m;
  struct gm_bsd_ether_pkt *pkt;
  int status, i;

  if (sc->open)
    {
      return GM_SUCCESS;
    }

  status =
    gm_open (&sc->port, sc->is->id, GM_ETHERNET_PORT_ID, "myri",
	     (enum gm_api_version) GM_API_VERSION);

  if (status != GM_SUCCESS)
    {
      GM_NOTE (("myri%d: error opening GM port,status=%d\n", sc->is_if.if_unit,
		status));
      status = EIO;
      goto abort_with_nothing;
    }

  sc->sreq = 0;
  sc->sdone = 0;
  sc->rdone = 0;

  gm_always_assert (GM_NUM_SEND_TOKENS >= GM_SEND_RING_CNT);
  gm_always_assert (GM_NUM_ETHERNET_RECV_TOKENS >= GM_RECV_RING_CNT);

  for (i = 0; i < GM_RECV_RING_CNT; i++)
    {
      pkt = sc->recv_ring + i;
      gm_always_assert (pkt->m == 0 && pkt->dma.nsegs == 0);
      m = gm_bsd_ether_arch_rxalloc (M_NOWAIT, GM_MAX_ETHERNET_SCATTER_CNT);
      if (m == NULL)
	{
	  status = ENOMEM;
	  goto abort_with_open_port;
	}
      pkt->m = m;
      if (!gm_bsd_ether_arch_encap (sc, m, &pkt->dma))
	{
	  status = ENOMEM;
	  goto abort_with_open_port;
	}
    }
  gm_pause_lanai (sc->is);
  gm_ethernet_set_recv_intr_callback (sc->port, gm_bsd_ether_rxeof, sc);
  gm_ethernet_set_sent_intr_callback (sc->port, gm_bsd_ether_txeof, sc);
  for (i = 0; i < GM_RECV_RING_CNT; i++)
    {
      pkt = sc->recv_ring + i;
      _gm_ethernet_provide_scatter_list (sc->port, pkt->dma.nsegs,
					 pkt->dma.desc);
    }
  gm_unpause_lanai (sc->is);
  sc->open = 1;
  return GM_SUCCESS;

 abort_with_open_port:
  gm_close (sc->port);
  for (i = 0; i < GM_RECV_RING_CNT; i++)
    {
      pkt = sc->recv_ring + i;
      if (pkt->dma.nsegs)
	{
	  gm_bsd_ether_arch_unencap (sc, &pkt->dma);
	}
      if (pkt->m)
	{
	  m_freem (pkt->m);
	  pkt->m = NULL;
	}
    }

 abort_with_nothing:

  return status;
}

int
gm_bsd_ether_close (struct gm_bsd_ether_sc *sc)
{

  struct gm_bsd_ether_pkt *pkt;
  int i;

  if (!sc->open)
    {
      return GM_SUCCESS;
    }

  gm_close (sc->port);
  for (i = 0; i < GM_RECV_RING_CNT; i++)
    {
      pkt = sc->recv_ring + i;
      gm_bsd_ether_arch_unencap (sc, &pkt->dma);
      m_freem (pkt->m);
      pkt->m = NULL;
    }
  sc->open = 0;
  return GM_SUCCESS;
}

/* trivial routines for checksum adding using linux names for compat
 * with the linux driver's csum offloading
 */
static gm_inline gm_u16_t
csum_fold (gm_u32_t sum)
{
  sum = (sum & 0xffff) + (sum >> 16);
  sum = (sum & 0xffff) + (sum >> 16);
  return ~sum;
}
static gm_inline gm_u32_t
csum_add (gm_u32_t sum)
{
  sum = (sum & 0xffff) + (sum >> 16);
  sum = (sum & 0xffff) + (sum >> 16);
  return sum;
}

/*
 * Adjust the hardware checksum given to us by the lanai so as to not
 * include the headers, then stash that in the packetheader so that
 * the IP stack can make use of it.
 *
 * Based heavily on Loic's gmip_skb_checksum() in the Linux driver.
 */

#if (GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED == 1)
void
gm_mbuf_rx_checksum (struct mbuf *top, u_int32_t length, u_int16_t csum)
{
  struct ether_header *eh;
  struct ip *ip;
  u_int16_t buf[8];
  u_int32_t hdr_csum, extra_csum, final_csum;
  int iphlen, real_packet_len, extra_bytes;


  eh = (struct ether_header *) (mtod (top, char *) + 2);

  if (csum == 0 || eh->ether_type != htons (ETHERTYPE_IP))
    {
      return;
    }

  /* card gave us an IP checksum of the message, and it is an IP packet */
  /* substract out the myri header, ethernet header and ip header */

  ip = (struct ip *) (sizeof (*eh) + mtod (top, char *) + 2);

  iphlen = (ip->ip_vhl & 0x0f) << 2;
  real_packet_len = 2 + sizeof (*eh) + ntohs (ip->ip_len);
  extra_bytes = length - real_packet_len;

  if (extra_bytes > sizeof (buf))
    {
      /* corrupt packet; don't smash the stack trying to csum these */
      /*  extra bytes */
      return;
    }

  /* internet checksum have "no endianess", so undo the ntohs done
     previously 
  */
  csum = htons (csum);

  top->m_pkthdr.len = length;

  hdr_csum = in_cksum (top, 2 + sizeof (*eh) + iphlen);
  /* un-invert the result */
  hdr_csum = ~hdr_csum & 0xffff;

  extra_csum = hdr_csum;

  if (extra_bytes > 0)
    {
      int i;

      bzero (buf, sizeof (buf));

      m_copydata (top, real_packet_len, extra_bytes, (char *) buf);
      if (extra_bytes & 1)
	{
	  u_int32_t last_extra_word = ntohs ((unsigned char) buf[0]);

	  extra_csum = csum_add (extra_csum + last_extra_word);
	  bzero (buf, sizeof (buf));
	  m_copydata (top, real_packet_len + 1, extra_bytes - 1, (char *) buf);
	}

      for (i = 0; i < extra_bytes / 2; i++)
	{
	  extra_csum = csum_add (extra_csum + buf[i]);
	}
    }

  final_csum = csum + csum_fold (extra_csum);
  final_csum = csum_add (final_csum);

#if (GM_OS_FREEBSD == 1)
  top->m_pkthdr.csum_data = htons (final_csum);
  top->m_pkthdr.csum_flags = CSUM_DATA_VALID;
#endif

#if (GM_OS_MACOSX == 1)
  top->m_pkthdr.csum_data = htons (final_csum ^ 0xffff);
  top->m_pkthdr.csum_flags =  CSUM_DATA_VALID | CSUM_TCP_SUM16;
#endif

#if (GM_OS_OSF1)
  csum = htons (final_csum);
  csum =
    in_pseudo (ip->ip_src.s_addr, ip->ip_dst.s_addr,
	       htonl (csum + ntohs (ip->ip_len) - iphlen + ip->ip_p));
  csum ^= 0xffff;
  if (csum == 0)
    {
      top->m_flags |= M_NOCHECKSUM;
    }
#endif
}

#endif /*GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED */

/*
 * Creates a new mbuf chain which is properly aligned & short enough
 * to fit in GM_MAX_ETHERNET_GATHER_CNT descriptors, and copies
 * the contents of the old chain to the new.
 */

struct mbuf *
gm_m_dup (struct mbuf *m_head)
{
  struct mbuf *m_src, *m_dst, *m_ret, *md_prev;
  char *src, *dst;
  int len, slen, dlen, copy_len, dused;

  m_ret = gm_bsd_ether_arch_rxalloc(M_NOWAIT, GM_MAX_ETHERNET_GATHER_CNT);
  if (m_ret == NULL)
    {
      return NULL;
    }
  m_dst = m_ret;
  m_src = m_head;

  slen = dlen = dused = 0;
  len = m_ret->m_pkthdr.len = m_head->m_pkthdr.len;
  while (len >= 0 && m_dst && m_src)
    {
      /* grab new pointers & lengths */
      if (slen == 0)
	{
	  src = mtod (m_src, char *);

	  slen = m_src->m_len;
	}
      if (dlen == 0)
	{
	  dst = mtod (m_dst, char *);

	  dlen = m_dst->m_len;
	  dused = 0;
	}
      /* copy the min of the data remaining the the src &
         dest. mbuf chains & the len  we need to copy */
      copy_len = (slen < dlen) ? slen : dlen;
      copy_len = (copy_len < len) ? copy_len : len;
      bcopy (src, dst, copy_len);
      /* increment pointers & decrement lengths */
      src += copy_len;
      dst += copy_len;
      slen -= copy_len;
      dlen -= copy_len;
      len -= copy_len;
      dused += copy_len;
      /* advance mbuf chain if needed */
      if (slen == 0)
	{
	  m_src = m_src->m_next;
	}
      if (dlen == 0)
	{
	  md_prev = m_dst;
	  m_dst = m_dst->m_next;
	}
    }

  if (len > 0)
    {
      GM_WARN (("ran out of data when copying packet for xmit\n"));
      m_freem (m_ret);
      return (NULL);
    }
  else
    {
      /* update the len field in the last dest mbuf to
         reflect how many bytes we're actually using */
      if (dlen != 0)
	{
	  m_dst->m_len = dused;
	  md_prev = m_dst;
	  m_dst = m_dst->m_next;
	}
      /* free the rest of the m_dst chain */
      if (m_dst)
	{
	  md_prev->m_next = NULL;
	  m_freem (m_dst);
	}
      return m_ret;
    }

}


/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
