
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */
#ifndef _gm_bsd_ether_h_
#define _gm_bsd_ether_h_

#define GM_MAX_ETHERNET_SG_CNT (GM_MAX_ETHERNET_SCATTER_CNT > GM_MAX_ETHERNET_GATHER_CNT ? GM_MAX_ETHERNET_SCATTER_CNT : GM_MAX_ETHERNET_GATHER_CNT)

#define GM_ETHER_MAXDMA		(GM_ETHERNET_MTU)
#define GM_SEND_RING_CNT	16	/* MUST be power of 2 */
#define GM_RECV_RING_CNT	GM_NUM_ETHERNET_RECV_TOKENS
#define GM_SEND_RING_MAX_INDEX (GM_SEND_RING_CNT -1)
#define GM_RECV_RING_MAX_INDEX (GM_RECV_RING_CNT -1)

#include "gm_arch.h"
#include "gm_bsd_ether_arch.h"

/*
 * This struct describes a pending send or recv. 
 */

struct gm_bsd_ether_pkt
{
  struct gm_bsd_ether_arch_dma_info dma;
  struct mbuf *m;
};

/*
 * The "softc" structure for the BSD ethernet driver.  (The term "softc"
 * is a standard BSD term for a driver instance and is commonly
 * abbreviated "sc")
 *	
 * Note that gm_bsd_ether_arch_sc arch contains platform dependant
 * data.  It must not be a pointer, and it must be the first field in
 * the softc struct.  This is because the network stack on some
 * platforms makes assumptions about the contents of the first part
 * of a network driver's softc
 */

struct gm_bsd_ether_sc
{  
  struct gm_bsd_ether_arch_sc arch;		/* must be 1st & not ptr */
  struct gm_instance_state *is;			/* gm instance */
  struct gm_port *port;				/* gm port */
  struct gm_bsd_ether_pkt send_ring[GM_SEND_RING_CNT];	/* pending sends */
  struct gm_bsd_ether_pkt recv_ring[GM_RECV_RING_CNT]; /* pending rcvs */
  int sreq;					/* #sends requested */
  int sdone;					/* #sends completed */
  unsigned rdone;				/* #receives */
  int open;					/* is device open?*/
};

/* functions provided by this file */
void gm_bsd_ether_print_mbuf (struct mbuf *m);

void gm_bsd_ether_start (struct ifnet *ifp);

void gm_bsd_ether_txeof (void *ptr);

void gm_bsd_ether_rxeof (void *ptr, unsigned int ulen, gm_u16_t csum);

int gm_bsd_ether_open (struct gm_bsd_ether_sc *sc);

int gm_bsd_ether_close (struct gm_bsd_ether_sc *sc);

#if (GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED == 1)
void gm_mbuf_rx_checksum (struct mbuf *, u_int32_t length, u_int16_t csum);
#endif

struct mbuf *gm_m_dup (struct mbuf *);

/* functions required by gm_bsd_ether.c, should be implemented by
   gm_<arch>_ether.c */

int gm_bsd_ether_arch_encap (struct gm_bsd_ether_sc *sc, 
			     struct mbuf *m_head, 
			     struct gm_bsd_ether_arch_dma_info *dma);

struct mbuf *gm_bsd_ether_arch_rxalloc (int how, int max_segs);

static gm_inline void gm_bsd_ether_arch_unencap (struct gm_bsd_ether_sc *sc,
					struct gm_bsd_ether_arch_dma_info *dma);

/* inlines which should be implemented in gm_bsd_ether_arch.h; may be
 * essentially noops for some platforms */

static gm_inline void gm_bsd_ether_arch_incr_tx_counters (struct gm_bsd_ether_arch_sc *sc,
							  struct mbuf *m);

static gm_inline void gm_bsd_ether_arch_incr_rx_counters (struct gm_bsd_ether_arch_sc *sc,
						 unsigned int ulen,
						 struct ether_header *eh);

static gm_inline void gm_bsd_ether_arch_sc_lock (struct gm_bsd_ether_arch_sc *sc);

static gm_inline void gm_bsd_ether_arch_sc_unlock (struct gm_bsd_ether_arch_sc *sc);

static gm_inline void gm_bsd_ether_arch_bpf_xmit (struct ifnet *ifp,
					 struct mbuf *m_head);

#endif /*_gm_bsd_ether_h_*/



/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
