/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

#if !GM_KERNEL
#include <stdlib.h>
#endif

#include "gm_call_trace.h"
#include "gm_compiler.h"
#include "gm_internal.h"
#include "gm_malloc_debug.h"
#include "gm_debug_malloc.h"

#if GM_DEBUG_MALLOC
struct gm_mdebug *gm_mdebug_root;
#endif

/* Return a pointer to the specified amount of memory.  In the kernel,
   the memory will be nonpageable. */

void *
_gm_malloc (unsigned long len)
{
  if (!len) 
    {
      return((void *)0);
    }

#if GM_KERNEL
  GM_CALLED ();
  GM_RETURN_PTR (__gm_arch_kernel_malloc (len, 0));
#else  /* !GM_KERNEL */
# undef malloc
  return malloc (len);
# define malloc GM_DO_NOT_USE (malloc)
#endif /* !GM_KERNEL */
}

GM_ENTRY_POINT
void *
gm_malloc (gm_size_t len)
{
  void *ret;

  GM_CALLED_WITH_ARGS ((GM_U64_TMPL, GM_U64_ARG (len)));
  
  ret = _gm_malloc (len);
  if (!ret)
    goto abort_with_nothing;
  
  if (GM_MDEBUG_RECORD_PTR (ret, gm_free) != GM_SUCCESS)
    goto abort_with_ret;
  
#if GM_KERNEL_MALLOC_ALIGNS_PAGES && GM_KERNEL
  gm_assert (GM_PAGE_ALIGNED (ret));
#endif

  GM_RETURN_PTR (ret);
  
 abort_with_ret:
  _gm_free (ret);
 abort_with_nothing:
  GM_RETURN_PTR (0);
}
