/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_chained_dma_h_
#define _gm_chained_dma_h_

#include "gm_bootstrap.h"
#include "gm_call_trace.h"
#include "gm_debug_chained_dma.h"
#include "gm_debug_lanai_dma.h"
#include "gm_dispatch.h"
#include "gm_dma_chain.h"
#include "gm_enable_debug_counters.h"
#include "gm_lanai_globals.h"
#include "gm_state_inlines.h"


extern int gm_dma_chain_ok (void);

/****************************************************************
 * Macros
 ****************************************************************/

/* start a simple RDMA into the recv queue */

#define start_GRANULAR_RDMA(lar, ear, len) do {				\
  gm_assert (GM_RDMA_ALIGNED (ear));					\
  gm_assert (GM_RDMA_ALIGNED (len));					\
  GM_DEBUG_CHAINED_DMA_ASSERT (NOTICED_NOT (RDMAING));			\
  gm_chained_dma_append (ear, lar, len,					\
			 (GM_DMA_DESCRIPTOR_L2E				\
			  + GM_DMA_DESCRIPTOR_WAKE));			\
} while (0)

/****************************************************************
 * Prototypes
 ****************************************************************/

gm_status_t gm_chained_dma_initialize (void);

/****************************************************************
 * Inlines
 ****************************************************************/

/* Return nonzero if there are no more incomplete DMAs.
   FOR DEBUGGING ONLY. */

#if GM_DEBUG || GM_DEBUG_CHAINED_DMA
static inline
int
gm_chained_dma_slow_all_done (void)
{
  gm_dma_descriptor_t *free_descriptor;
  gm_dma_descriptor_t *last;

  free_descriptor = gm_chained_dma_free_descriptor;
  ;
  ;
  gm_assert (GM_ALIGNED (free_descriptor, 8));
  last = free_descriptor - 1;
  if (last < &gm.user_dma_descriptor[0])
    {
      last = &gm.user_dma_descriptor[GM_NUM_ELEM(gm.user_dma_descriptor)-1];
    }
  return ((gm_size_t) last->next_with_flags & GM_DMA_DESCRIPTOR_TERMINAL) != 0;
}
#endif /* GM_DEBUG */

extern int gm_chained_dma_initialized;
extern gm_dma_descriptor_t *gm_chained_dma_free_descriptor;

/* Append a DMA to the list of chained DMAs.

   Valid flags are
   GM_DMA_DESCRIPTOR_WAKE: generate a DMA_INT_BIT after this DMA segment
     is done.
   GM_DMA_DESCRIPTOR_E2L: DMA from ebus to lbus (SDMA direction)
   GM_DMA_DESCRIPTOR_L2E: DMA from lbus to ebus (RDMA direction) */

#ifdef GM_STATE_MACHINE
static inline
void
gm_chained_dma_append (gm_dp_t dp, gm_lp_t lp, gm_u32_t len, int flags)
{
  gm_dma_descriptor_t *free_descriptor, *next;
  gm_lp_n_t next_with_flags;
  
  GM_CALLED_WITH_ARGS ((GM_U64_TMPL",%p,%d,0x%x",
			GM_U64_ARG (dp), lp, len, flags));
  gm_assert (gm_chained_dma_initialized);
  gm_assert (len);
  gm_assert ((flags & GM_DMA_DESCRIPTOR_TERMINAL) == 0);
  gm_assert (gm_noticed_dma_addr (&gm.debug.dma_page_bitmap, dp)
	     == GM_SUCCESS);
  
  free_descriptor = gm_chained_dma_free_descriptor;
  GM_DEBUG_CHAINED_DMA_ASSERT (GM_ALIGNED (free_descriptor, 8));
  
  LOG_DISPATCH (0, "Appending a descriptor to DMA chain");

  if (flags & GM_DMA_DESCRIPTOR_WAKE)
    {
      GM_INCR_DEBUG_CNT (gm.outstanding_dma_cnt);
      if (GM_ENABLE_DEBUG_COUNTERS)
	{
	  GM_DEBUG_CHAINED_DMA_ASSERT (gm.outstanding_dma_cnt_debug_cnt <= 2);
	}
      
      /* Update software state to indicate that the DMA is not finished,
	 and to allow us to determine when it is finished. */

      GM_DEBUG_CHAINED_DMA_ASSERT (NOTICED (GM_STATE_MACHINE_NOTICE_NOT));
      NOTICE_NOT (GM_STATE_MACHINE_NOTICE_NOT);
      gm.GM_STATE_MACHINE.tail_dma_descriptor = free_descriptor;

      /* HACK: Infer which state machine is DMAing.  This inference
	 will not be valid in GM-2, where gn_get() can cause the RDMA
	 engine to do an SDMA. */

      if (flags & GM_DMA_DESCRIPTOR_E2L)
	{
	  GM_DEBUG_CHAINED_DMA_ASSERT (NOTICED_NOT (SDMAING));
	  NOTICE (SDMAING);
	}
      else
	{
	  GM_DEBUG_CHAINED_DMA_ASSERT (NOTICED_NOT (RDMAING));
	  NOTICE (RDMAING);
	}
    }
  
  /* Append the DMA to the chain */
  
  next_with_flags = free_descriptor->next_with_flags;
  GM_DEBUG_CHAINED_DMA_ASSERT ((gm_size_t) next_with_flags
			       & GM_DMA_DESCRIPTOR_TERMINAL);
  free_descriptor->len = len;
  free_descriptor->lanai_addr = lp;
  if (sizeof (dp) == 8)
    {
      free_descriptor->ebus_addr_high = (gm_u32_t) ((dp >> 31) >> 1);
    }
  else
    {
      gm_assert (free_descriptor->ebus_addr_high == 0);
    }
  GM_STBAR ();
  free_descriptor->ebus_addr_low = (gm_u32_t) dp;
  next = GM_DMA_DESCRIPTOR_NEXT_WITH_FLAGS_PTR (next_with_flags);
  gm_assert (next == free_descriptor + 1
	     || next == &gm.user_dma_descriptor[0]);
  gm_assert ((gm_size_t) next->next_with_flags & GM_DMA_DESCRIPTOR_TERMINAL);
  gm_chained_dma_free_descriptor = next;
  GM_STBAR ();
  free_descriptor->next_with_flags
    = GM_DMA_DESCRIPTOR_BUILD_NEXT_WITH_FLAGS (next, flags);

  if (GM_TRACE_LANAI_DMA)
    {
      GM_PRINT (1, ("chained descriptor:\n"));
      gm_hex_dump (free_descriptor, sizeof (*free_descriptor));
    }

  /* Make sure the FPGA notices the appended segment descriptor */

  gm_write_PULSE (4);

  /* Handle IO caching issues on Sparcs.  This may append a single DMA
     to the previously enqueued DMA. */

  if (GM_SPARC_STREAMING)
    {
      gm_dp_t limit;
      
      limit = dp + len;
      if (((flags & GM_DMA_DESCRIPTOR_DIR) == GM_DMA_DESCRIPTOR_L2E
	   && !GM_ALIGNED (limit, GM_SPARC_CACHE_LINE))
	  || ((flags & GM_DMA_DESCRIPTOR_DIR) == GM_DMA_DESCRIPTOR_E2L
	      && !GM_ALIGNED (limit, GM_SPARC_IO_PAGE_LEN)))
	{
	  /* Either the end of the RDMA is not cache line aligned, or the
	     last byte of the SDMA is not in the last cache line of a
	     page.  In the RDMA case, we need to do an SDMA readback in
	     the same page to flush the RDMA.  In the SDMA case, we want
	     to read from the last cache line in the page to leave the
	     prefetch cache in the a known state of having read from the
	     last cache line in the page.  In both cases, we use the DMA
	     chaining hardware to schedule a small SDMA from the very end
	     of the page after the desired DMA. */

	  next_with_flags = next->next_with_flags;
	  free_descriptor = next;
	  ;
	  next = GM_DMA_DESCRIPTOR_NEXT_WITH_FLAGS_PTR (next_with_flags);
      
	  if (sizeof (limit) == 8)
	    {
	      free_descriptor->ebus_addr_high = (limit >> 31) >> 1;
	    }
	  else
	    {
	      gm_assert (free_descriptor->ebus_addr_high == 0);
	    }
	  free_descriptor->len = 2 * GM_SPARC_STREAMING_FLUSH_LEN;
	  free_descriptor->lanai_addr = (gm_lp_t)gm.dummy_flush_target;
	  /* read at end of page */
	  free_descriptor->ebus_addr_low
	    = (GM_ALIGN (u32, limit, GM_SPARC_IO_PAGE_LEN)
	       + (GM_SPARC_IO_PAGE_LEN 
		  - GM_SPARC_STREAMING_FLUSH_LEN 
		  - GM_SPARC_CACHE_LINE));
	  free_descriptor->next_with_flags
	    = (GM_DMA_DESCRIPTOR_BUILD_NEXT_WITH_FLAGS
	       (next, GM_DMA_DESCRIPTOR_E2L));

	  gm_write_PULSE (4);
	}
    }
  
  GM_RETURN_NOTHING ();
}
#endif /* GM_STATE_MACHINE */

/* Spin until no DMA is in progress. */

static inline void
await_user_DMA_completion (void)
{
  /* wait for all DMAs to complete */
  
  while (!user_DMA_done ())
    ;

  /* Since all DMAs are done, infer SDMA_INT_BIT and RDMA_INT_BIT, because
     it can be expensive to compute these elsewhere, but is very cheap here. */
  
  if (NOTICED (SDMAING) && NOTICED_NOT (SDMA_INT_BIT))
    {
      GM_DEBUG_CHAINED_DMA_ASSERT
	(gm.sdma.tail_dma_descriptor != &gm.dma.fake_terminal_dma_descriptor);
      GM_DEBUG_CHAINED_DMA_ASSERT
	(gm.sdma.tail_dma_descriptor >= &gm.user_dma_descriptor[0]);
      GM_DEBUG_CHAINED_DMA_ASSERT
	(gm.sdma.tail_dma_descriptor
	 <= &gm.user_dma_descriptor[GM_NUM_ELEM (gm.user_dma_descriptor) - 1]);
      GM_DEBUG_CHAINED_DMA_ASSERT
	(GM_DMA_DESCRIPTOR_TERMINAL
	 & (GM_DMA_DESCRIPTOR_NEXT_WITH_FLAGS_FLAGS
	  (gm.sdma.tail_dma_descriptor->next_with_flags)));
      
      gm.sdma.tail_dma_descriptor = &gm.dma.fake_terminal_dma_descriptor;
      gm_assert (NOTICED_NOT (SDMA_INT_BIT));
      NOTICE (SDMA_INT_BIT);
      GM_DECR_DEBUG_CNT (gm.outstanding_dma_cnt);
      GM_DEBUG_CHAINED_DMA_ASSERT (get_ISR () & DMA_INT_BIT);
      gm_clear_DMA_INT_BIT ();
      /* GM_DEBUG_CHAINED_DMA_ASSERT ((get_ISR () & DMA_INT_BIT) == 0); */
    }
  
  if (NOTICED (RDMAING) && NOTICED_NOT (RDMA_INT_BIT))
    {
      GM_DEBUG_CHAINED_DMA_ASSERT
	(gm.rdma.tail_dma_descriptor != &gm.dma.fake_terminal_dma_descriptor);
      GM_DEBUG_CHAINED_DMA_ASSERT
	(gm.rdma.tail_dma_descriptor >= &gm.user_dma_descriptor[0]);
      GM_DEBUG_CHAINED_DMA_ASSERT
	(gm.rdma.tail_dma_descriptor
	 <= &gm.user_dma_descriptor[GM_NUM_ELEM (gm.user_dma_descriptor) - 1]);
      GM_DEBUG_CHAINED_DMA_ASSERT
	(GM_DMA_DESCRIPTOR_TERMINAL
	 & (GM_DMA_DESCRIPTOR_NEXT_WITH_FLAGS_FLAGS
	  (gm.rdma.tail_dma_descriptor->next_with_flags)));
      
      gm.rdma.tail_dma_descriptor = &gm.dma.fake_terminal_dma_descriptor;
      gm_assert (NOTICED_NOT (RDMA_INT_BIT));
      NOTICE (RDMA_INT_BIT);
      GM_DECR_DEBUG_CNT (gm.outstanding_dma_cnt);

      /* For RDMAs, the wake bit can be arbitrarily delayed until the
	 host accepts the all data from the PCIDMA chip's FIFO.
	 Therefore, we read from the PCI bus to ensure that any
	 WAKE corresponding to the DMAs has completed. */

      synchronous_RDMA (&gm.trash, gm.page_hash.bogus_sdma_ptr, 4);

      gm_clear_DMA_INT_BIT ();
      /* GM_DEBUG_CHAINED_DMA_ASSERT ((get_ISR () & DMA_INT_BIT) == 0); */
    }
}

/* return the ethernet checksum for a just-completed DMA or 0 if
   there is none. */

static inline gm_u16_t
rdma_checksum (void)
{
  int done;

  /* ensure the DMA descriptor writeback has occurred. */
  do 
    {
      done = (GM_DMA_DESCRIPTOR_TERMINAL 
	      & (gm_size_t)gm.ethernet.recv.checksum_descriptor->next_with_flags);
    }
  while (!done);

#if (L7 || L9) && !defined CSPI_MAP26xx
  return gm.ethernet.recv.checksum_descriptor->csum[1];
#elif L6
  {
    gm_u32_t a;

    a = gm.ethernet.recv.checksum_descriptor->csum[0];
    a += gm.ethernet.recv.checksum_descriptor->csum[1];
    a += a >> 16;		/* add back any 16-bit carry. */
    return (gm_u16_t) (0xffff & a);
  }
#else
  return 0;
#endif
}

#endif /* _gm_chained_dma_h_ */


/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
