/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_connection_h_
#define _gm_connection_h_

/****************************************************************
 * Switches
 ****************************************************************/

#define GM_DEBUG_CONNECTIONS 0

/****************************************************************
 * Includes
 ****************************************************************/

#include "gm_rand.h"
#include "gm_send_tokens.h"
#include "gm_types.h"
#include "gmcp.h"

/****************************************************************
 * connection ack/nack functions
 ****************************************************************/

/* Special _gm_reack() entry point to pipeline gm_ack() */

static inline void
_gm_reack (gm_connection_t * c, unsigned int c__ack_pending)
{
  gm_s32_t rtc;

  rtc = RTC;
  gm_assert (&gm_connection[0] - c <= 0 &&
	     &gm_connection[GM_NUM_PRIORITIES * (gm.max_node_id + 1)] - c >
	     0);
  if (!c__ack_pending)
    {
      c->next_to_ack = gm.first_connection_to_ack;
      gm.first_connection_to_ack = c;
      c->ack_pending = -1;
    }
  NOTICE (ACK_PENDING);
  c->known_alive_time = rtc;
}

/* Cause an ack for a connection to be scheduled for sending to the
   remote end of the connection. */

static inline void
gm_reack (gm_connection_t * c)
{
  _gm_reack (c, c->ack_pending);
}

/* Ack a message received on a connection. */

static inline void
gm_ack (gm_connection_t * c)
{
  unsigned int c__ack_packet_sexno_parts_seqno;
  unsigned int c__ack_pending;

  c__ack_packet_sexno_parts_seqno = c->ack_packet.sexno.parts.seqno;
  c__ack_pending = c->ack_pending;
  c->ack_packet.subtype = GM_ACK_SUBTYPE;
  c->ack_packet.usecs_delay = 0;
  c->ack_packet.sexno.parts.seqno = c__ack_packet_sexno_parts_seqno + 1;
  _gm_reack (c, c__ack_pending);
}

/**********************************************************************/

/* Note: It is up to the caller to restore the send_sexno for the
   connection. */
static inline void
rewind_connection (gm_connection_t * c)
{
  gm_send_record_t *sent;

  sent = c->first_send_record;
  if (!sent)
    return;

  /* fairly allow the next subport to be the first to send. */
  /* BAD: should be pipelined with preceeding code */

  c->first_active_send_port = sent->send_token->common.subport->next;

  rewind_send_tokens (sent);

  /* Free the send records */
  c->last_send_record->next = gm.free_send_records;
  gm.free_send_records = c->first_send_record;
  c->first_send_record = 0;
}

void _gm_connection_init (gm_connection_t * c);
void gm_connection_reset (gm_connection_t * c);

#endif /* _gm_connection_h_ */
