/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1996, 1997, 1998, 1999, 2000 by Myricom, Inc.		 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_dispatch_h_
#define _gm_dispatch_h_

#include "gm_bootstrap.h"
#include "gm_call_trace.h"
#include "gm_chained_dma.h"
#include "gm_compiler.h"
#include "gm_dispatch_logging.h"
#include "gm_lanai_def.h"
#include "gm_lanai_globals.h"
#include "gm_lzero.h"
#include "gm_state.h"
#include "gm_types.h"

/****************************************************************
 * Macros
 ****************************************************************/

extern int state_modified_since_dispatch_offset;

/****************
 * Handler entry logging support.
 ****************/

#define GM_HANDLER_CALLED_LABEL invoke_GM_HANDLER_CALLED_first

/* Record the current handler.  This is split into a separate macro
   to suppress the emission of the __GM_FUNCTION__ string into the
   code when not debugging. */

#if GM_DEBUG || GM_ENABLE_LOG_DISPATCH
#define GM_RECORD_CURRENT_HANDLER() do {				\
  gm.current_handler = __GM_FUNCTION__;					\
  gm.current_handler_extra = "";					\
} while (0)
#else
#define GM_RECORD_CURRENT_HANDLER()
#endif

/* Notice that a handler is called.  This macro should be called
   at the start of every firmware dispatch handler. */

#define GM_HANDLER_CALLED() do {					\
 GM_HANDLER_CALLED_LABEL:						\
  GM_CALLED ();								\
  GM_RECORD_CURRENT_HANDLER ();						\
} while (0)

/* Macro to cause a compilation error if GM_HANDLER_CALLED() is not
   been invoked in the current function.  This is used inside
   DISPATCH() to ensure that every function that dispatches (i.e.:
   ever GM dispatch handler) has GM_HANDLER_CALLED() invoked in the
   handler. */

#define GM_REQUIRE_GM_HANDLER_CALLED() do {				\
  if (0)								\
    {									\
      goto GM_HANDLER_CALLED_LABEL;					\
    }									\
} while (0)

/****************
 * DISPATCH support
 ****************/

/* Checks to perform in every dispatch, no matter how the dispatch is
   implemented. */

#define GM_EVERY_DISPATCH() do {					\
  GM_REQUIRE_GM_HANDLER_CALLED ();					\
									\
  if (GM_DEBUG_CHAINED_DMA						\
      && NOTICED (SDMAING)						\
      && NOTICED_NOT (SDMA_INT_BIT))					\
    {									\
      GM_DEBUG_CHAINED_DMA_ASSERT					\
	(gm.sdma.tail_dma_descriptor >= &gm.user_dma_descriptor[0]);	\
      GM_DEBUG_CHAINED_DMA_ASSERT					\
	(gm.sdma.tail_dma_descriptor					\
	 <= &(gm.user_dma_descriptor					\
	      [GM_NUM_ELEM (gm.user_dma_descriptor) - 1]));		\
    }									\
  if (GM_DEBUG_CHAINED_DMA						\
      && NOTICED (RDMAING)						\
      && NOTICED_NOT (RDMA_INT_BIT))					\
    {									\
      GM_DEBUG_CHAINED_DMA_ASSERT					\
	(gm.rdma.tail_dma_descriptor >= &gm.user_dma_descriptor[0]);	\
      GM_DEBUG_CHAINED_DMA_ASSERT					\
	(gm.rdma.tail_dma_descriptor					\
	 <= &(gm.user_dma_descriptor					\
	      [GM_NUM_ELEM (gm.user_dma_descriptor) - 1]));		\
    }									\
  if (GM_ENABLE_DEBUG_COUNTERS)						\
    {									\
      GM_DEBUG_CHAINED_DMA_ASSERT					\
	(gm.outstanding_dma_cnt_debug_cnt <= 2);			\
      GM_DEBUG_CHAINED_DMA_ASSERT					\
	(((NOTICED (SDMAING) && NOTICED_NOT (SDMA_INT_BIT)		\
	   ? 1U								\
	   : 0U)							\
	  + (NOTICED (RDMAING) && NOTICED_NOT (RDMA_INT_BIT)		\
	     ? 1U							\
	     : 0U))							\
	 == gm.outstanding_dma_cnt_debug_cnt);				\
    }									\
} while (0)

typedef void (*gm_handler_t) (void);

/* Got a handler whose pointer has been prefetched.  This is used in
   some handlers where we wish to schedule instructions in the dispatch
   delay slots. */

#define GOTO_PREFETCHED_HANDLER(handler) do {			\
  gm_handler_t GOTO_PREFETCHED_HANDLER_handler;			\
								\
  gm_assert (!isr_state_modified_since_get_isr);		\
  gm_assert (!state_modified_since_dispatch_offset);		\
  GM_EVERY_DISPATCH ();						\
  GM_RECORD_RETURN ("%s", "");					\
  asm volatile ("mov %1,%?pc\n\t"				\
		"sub %?fp,8,%?sp\n\t"				\
		"nop"						\
		: "=r" (GOTO_PREFETCHED_HANDLER_handler)	\
		: "r" (handler)					\
		: "fp", "sp", "pc");				\
  /* never get here */						\
  gm_abort ();							\
} while (0)

/* GOTO_HANDLER_AT_OFFSET(): Goto a handler corresponding to a
   given event index offset.  This is used in some handler where we
   wish to schedule instructions in the dispatch delay slots. */
   
#define _GOTO_HANDLER_AT_OFFSET(offset) do {	\
  gm_handler_t GOTO_HANDLER_handler;		\
						\
  GM_RECORD_RETURN ("%s\n", "");		\
  asm volatile ("ld [%1 add %2],%?pc\n\t"	\
		"sub %?fp,8,%?sp\n\t"		\
		"nop\n\t"			\
		"nop\n\t"			\
		"nop"				\
		: "=r" (GOTO_HANDLER_handler)	\
		: "r" (gmptr), "r" (offset)	\
		: "fp", "sp", "pc");		\
  /* never get here */				\
  gm_abort ();					\
} while (0)

#define GOTO_HANDLER_AT_OFFSET(offset) do {				\
  gm_assert (!isr_state_modified_since_get_isr);			\
  gm_assert (!state_modified_since_dispatch_offset);			\
  GM_EVERY_DISPATCH ();							\
  _GOTO_HANDLER_AT_OFFSET (offset);					\
} while (0)

/* Compute the handler pointer given the C function pointer of the
   handler.  This is necessary because we are going to use an
   optimized "goto" to switch to the new handler skipping.  This skips
   the first 2 instructions (8 bytes) of the function preamble,
   hence the "+ 8". */

#define GM_REFERENCE_LABEL(label) ((gm_handler_t) ((char *) label + 8))

				/* Special Macro to efficiently set the
				   handler in the dispatch table. */
#define SET_HANDLER(event_type, label, label_extra) do {	\
  gm.handler[event_type]					\
    = GM_REFERENCE_LABEL (GM_CAT (label, label_extra));		\
} while (0)

#define GET_HANDLER(event_type) (gm.handler[event_type])

				/* Macro to check a handler in the dispatch
				   table. */
#define ASSERT_HANDLER(event_type, label, label_extra) do {		\
  gm_assert (gm.handler[event_type]					\
	       == GM_REFERENCE_LABEL (GM_CAT (label,label_extra)));	\
} while (0)

static inline void
check_handler (unsigned gm_offset)
{
  if (GM_DEBUG && (gm_offset & 3))
    {
      gm_printf ("Error (1): offset = 0x%x\n", gm_offset);
      fflush (stdout);
      gm_always_assert (0);
    }
  if (GM_DEBUG && (gm_offset < 16 || gm_offset >= 16 + NUM_EVENT_TYPES * 4))
    {
      gm_printf ("Error (2): offset = 0x%x\n", gm_offset);
      fflush (stdout);
      gm_always_assert (0);
    }
}

/* check for stack leaks */

#if !GM_DEBUG
#define check_ps(unique)
#else  /* GM_DEBUG */
#define check_ps(unique) do {						\
  char *fp;								\
  asm ("mov %?fp,%0\n\t" "nop\n\t" "nop":"=r" (fp));			\
  if (!gm.debug.old_fp)							\
    {									\
      gm.debug.old_fp = fp;						\
    }									\
  /* maybe this function is inlined */					\
  if (fp != gm.debug.old_fp)						\
    {									\
      gm_printf ("fp = %p, old_fp = %p\n", fp, gm.debug.old_fp);	\
      gm_printf ("unique = %d\n", (int) unique);			\
      fflush (stdout);							\
      gm_abort ();							\
    }									\
  gm_assert (fp >= (char *) gm_stack);					\
  gm_assert (fp <= (char *) gm_top_of_stack);				\
  gm_assert (gm.send_chunk[0].cafebabe == 0xcafebabe);			\
  gm_assert (gm.send_chunk[1].cafebabe == 0xcafebabe);			\
  gm_assert (gm.recv_chunk[0].cafebabe == 0xcafebabe);			\
  gm_assert (gm.recv_chunk[1].cafebabe == 0xcafebabe);			\
  gm_assert (gm_stack[0] == 0);						\
  gm_assert (gm_stack[1] == 0);						\
  gm_assert (gm_stack[2] == 0);						\
  gm_assert (gm_stack[3] == 0);						\
} while (0)
#endif /* GM_DEBUG */

/****************
 * _DISPATCH()
 ****************/

/* This tiny pair of macros is actually a major HACK.  It requires
   that the _gm_event_index is just below the "gm" lanai globals
   structure with the entries in the table reversed so that they can
   be referenced by reverse-indexing off the "gm" pointer.
   Furthermore, the entries in this gm_event_index table must be
   computed using GM_SET_EVENT_INDEX_FOR_STATE (index, state) which
   stores the precomputed offset from "&gm" to the handler pointer to
   use, which furthermore requires that all the entries in the
   gm.handler table start no more than 255 bytes from the start of the
   gm structure.

   The point of all this is that it allows the ubiquitous "&gm"
   pointer that the compiler keeps in a register to be used to access
   these tables optimally efficiently, allowing for a 7-insn dispatch
   on the LANai4. */

extern unsigned char _gm_event_index[BIGGEST_STATE_BIT << 1];

static inline gm_u32_t
DISPATCH_OFFSET (gm_u32_t isr)
{
  gm_u32_t imr;
  
  /* We do not use the get_IMR() macro here, because we wish to avoid the
     extra safety barriers, and we know it is safe. --Glenn */

#if L4 | L5 | L6
  imr = IMR;
#elif L7 | L8 | L9
  imr = IMR_INIT_VAL;
#else
#error unrecognized LANai value
#endif

  if (GM_DEBUG)
    {
      state_modified_since_dispatch_offset = 0;
    }

  return ((unsigned char *) ((gm_u32_t) gmptr - GM_STATE))[-((isr & imr))];
}

#define _DISPATCH(unique, msg, base_msg) do {				\
  gm_u32_t _offset;							\
  log_it (unique, msg, LZERO, base_msg);				\
  check_ps (unique);							\
  _offset = DISPATCH_OFFSET(get_ISR());	                                \
  GM_LOG_EVT(_offset);							\
  check_handler(_offset);						\
  GOTO_HANDLER_AT_OFFSET (_offset);					\
} while (0)

/* L7 function return:
	ld -4[%fp],%pc  			SKIP
	ld -8[%fp],%fp  			COMBINEv
	add %fp,0x0,%sp				COMBINE^ ld -8[fp],%sp
	nop
	nop
   L7 function call:	
	mov _check_handler,%pc
	mov .+8,%rca				SKIP
	st %rca,[--%sp]				SKIP
   L7 prologue:
	st %fp,-4[*%sp]	!push old FP		SKIP
	add %sp,8,%fp	!generate new FP 	ELIMINABLE???

	
	ld -4[%fp],%sp
	mov HANDLER,%pc
	nop 
	add %sp,8,%fp
*/

/****************
 * DISPATCH()
 ****************/

#if GM_DEBUG || GM_ENABLE_LOG_DISPATCH
#define DISPATCH(unique, msg) do {					\
  _DISPATCH (unique, __GM_FILE__ ":" __GM_LINE__ ": " msg, msg);	\
} while (0)
#else
/* This version prevents unneeded strings from being emitted by the
   compiler. */
#define DISPATCH(unique, msg) do {					\
  _DISPATCH (unique, 0, 0);						\
} while (0)
#endif

#ifndef INITIAL_DISPATCH
#define INITIAL_DISPATCH(a, b) DISPATCH (a, b)
#endif

#define GM_CHAIN_DISPATCH(unique, msg, event, handler) do {		\
  LOG_DISPATCH (unique, msg);						\
  SET_HANDLER (event, handler);						\
  if (GM_DEBUG)								\
    {									\
      isr_state_modified_since_get_isr = 0;				\
      state_modified_since_dispatch_offset = 0;				\
    }									\
  GOTO_PREFETCHED_HANDLER (GM_REFERENCE_LABEL (handler));		\
} while (0)

/****************
 * GM_SET_EVENT_INDEX_FOR_STATE()
 ****************/

#define GM_SET_EVENT_INDEX_FOR_STATE(i,s) do {				\
  gm_assert (GM_OFFSETOF (gm_lanai_globals_t, handler[(i)]) <= 255);	\
  ((unsigned char *) gmptr)[-(s)]					\
    = GM_OFFSETOF (gm_lanai_globals_t, handler[(i)]);			\
} while (0)

/****************
 * print_state
 ****************/

/****************************************************************
 * Handler declarations
 ****************************************************************/

void dma__handle_dma_interrupt (void);
void dma__handle_rdma_interrupt (void);
void dma__handle_sdma_interrupt (void);
void dma__panic (void);
void fair__rdma_0 (void);
void fair__rdma_1 (void);
void fair__sdma_0 (void);
void fair__sdma_1 (void);
void fair__sdma_rdma_0 (void);
void fair__sdma_rdma_1 (void);
void fair__sdma_rdma_2 (void);
void idle (void);
void rdma__chunk_done (void);
void rdma__continue_chunk (void);
void rdma__continue_last_chunk_0 (void);
void rdma__continue_last_chunk_1 (void);
void rdma__last_chunk_done_0 (void);
void rdma__last_chunk_done_1 (void);
void rdma__maybe_finish_ethernet_rdma (void);
void rdma__rdma_chunk_0 (void);
void rdma__rdma_chunk_1 (void);
void rdma__token_done (void);
void recv__discard_overflow_0 (void);
void recv__discard_overflow_1 (void);
void recv__done_discarding_overflow_0 (void);
void recv__done_discarding_overflow_1 (void);
void recv__got_chunk_0 (void);
void recv__got_chunk_1 (void);
void recv__start_receiving_chunk_0 (void);
void recv__start_receiving_chunk_1 (void);
void sdma__continue_sdma (void);
void sdma__finish_ethernet_sdma (void);
void sdma__finish_sdma (void);
void sdma__poll_for_sdma_0 (void);
void sdma__start_sdma_0 (void);
void sdma__start_sdma_1 (void);
void send__finish_sending_chunk (void);
void send__start_sending_ack (void);
void send__start_sending_chunk_0 (void);
void send__start_sending_chunk_1 (void);
void timer (void);
void timer__update_host_sent_queues__rdma (void);
void timer__update_host_sent_queues__sdma (void);

#endif /* _gm_dispatch_h_ */
