/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_dispatch.h"
#include "gm_chained_dma.h"
#include "gm_types.h"
#include "gm_enable_debug_counters.h"

/* Handle getting a DMA_INT_BIT when only an SDMA is in progress. */

void
dma__handle_sdma_interrupt ()
{
  GM_HANDLER_CALLED ();

  GM_DEBUG_CHAINED_DMA_ASSERT (get_ISR () & DMA_INT_BIT);
  gm_clear_DMA_INT_BIT ();
  /* gm_assert ((get_ISR () & DMA_INT_BIT) == 0); */

  gm_assert
    (gm.rdma.tail_dma_descriptor == &gm.dma.fake_terminal_dma_descriptor);
  gm_assert
    (gm.sdma.tail_dma_descriptor != &gm.dma.fake_terminal_dma_descriptor);

  /* Make sure the DMA really completed */
  GM_DEBUG_CHAINED_DMA_ASSERT
    (GM_DMA_DESCRIPTOR_TERMINAL
      & (GM_DMA_DESCRIPTOR_NEXT_WITH_FLAGS_FLAGS
	 (gm.sdma.tail_dma_descriptor->next_with_flags)));
  
  gm.sdma.tail_dma_descriptor = &gm.dma.fake_terminal_dma_descriptor;
  gm_assert (NOTICED_NOT (SDMA_INT_BIT));
  NOTICE (SDMA_INT_BIT);
  GM_DECR_DEBUG_CNT (gm.outstanding_dma_cnt);
  DISPATCH (1, "Handled SDMA_INT_BIT");
}

/* Handle getting a DMA_INT_BIT when only an RDMA is in progress. */

void
dma__handle_rdma_interrupt ()
{
  GM_HANDLER_CALLED ();

  GM_DEBUG_CHAINED_DMA_ASSERT (get_ISR () & DMA_INT_BIT);
  gm_clear_DMA_INT_BIT ();
  /* gm_assert ((get_ISR () & DMA_INT_BIT) == 0); */

  gm_assert
    (gm.sdma.tail_dma_descriptor == &gm.dma.fake_terminal_dma_descriptor);
  gm_assert
    (gm.rdma.tail_dma_descriptor != &gm.dma.fake_terminal_dma_descriptor);

  /* Make sure the DMA really completed */
  GM_DEBUG_CHAINED_DMA_ASSERT
    (GM_DMA_DESCRIPTOR_TERMINAL
      & (GM_DMA_DESCRIPTOR_NEXT_WITH_FLAGS_FLAGS
	 (gm.rdma.tail_dma_descriptor->next_with_flags)));
  
  gm.rdma.tail_dma_descriptor = &gm.dma.fake_terminal_dma_descriptor;
  gm_assert (NOTICED_NOT (RDMA_INT_BIT));
  NOTICE (RDMA_INT_BIT);
  GM_DECR_DEBUG_CNT (gm.outstanding_dma_cnt);
  DISPATCH (2, "Handled RDMA_INT_BIT");
}

/* Handle getting a DMA_INT_BIT when no DMA is in progess.  This
   should never happen. */

void
dma__panic ()
{
  GM_HANDLER_CALLED ();
  gm_write_PULSE (6);
  gm_always_assert (!"bogus DMA state after pulses");
  DISPATCH (3, "never get here.");
}

/* Handle getting a DMA_INT_BIT when both and SDMA and an RDMA are in
   progress.  This is very costly. */

void
dma__handle_dma_interrupt ()
{
  gm_dma_descriptor_t *sdma, *rdma;
  gm_size_t sdma__next_with_flags, rdma__next_with_flags;

  GM_HANDLER_CALLED ();

  /****************
   * This does not work for unknown reasons.  None of the HW folks can
   * explain why.
   ****************/

  /* Clear the hardware DMA interrupt bit.  Since the second DMA may be
     interrupting for an entire PCI clock cycle here, we must clear the
     bit twice at least 6 cycles appart (200MHz/33MHz) */
  GM_DEBUG_CHAINED_DMA_ASSERT (get_ISR () & DMA_INT_BIT);
  gm_clear_DMA_INT_BIT ();

  /* Get pointers to the tail descriptors of any ongoing DMAs. */
  
  sdma = gm.sdma.tail_dma_descriptor;
  rdma = gm.rdma.tail_dma_descriptor;

  gm_assert (sdma != &gm.dma.fake_terminal_dma_descriptor);
  gm_assert (rdma != &gm.dma.fake_terminal_dma_descriptor);
  
  /* Wait long enough to ensure that any descriptor corresponding to
     the DMA_INT_BIT has been updated. */

  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");
  asm volatile ("nop\n\tnop\n\tnop\n\tnop\n\tnop");

  /* Read the "terminal" bit from each descriptor. */
  
  sdma__next_with_flags = (gm_size_t) sdma->next_with_flags;
  rdma__next_with_flags = (gm_size_t) rdma->next_with_flags;
  if (sdma__next_with_flags & GM_DMA_DESCRIPTOR_TERMINAL)
    {
      /* The pending SDMA completed, so handle it. */
      
      gm_assert
	(gm.dma.fake_terminal_dma_descriptor.next_with_flags
	 == (gm_lp_n_t) GM_DMA_DESCRIPTOR_TERMINAL);
      gm.sdma.tail_dma_descriptor = &gm.dma.fake_terminal_dma_descriptor;
      NOTICE (SDMA_INT_BIT);
      GM_DECR_DEBUG_CNT (gm.outstanding_dma_cnt);

      /* Check if the RDMA also completed. */
      
      if (rdma__next_with_flags & GM_DMA_DESCRIPTOR_TERMINAL)
	{
	  /* The pending RDMA completed, so handle it. */
	  
	  gm_assert
	    (gm.dma.fake_terminal_dma_descriptor.next_with_flags
	     == (gm_lp_n_t) GM_DMA_DESCRIPTOR_TERMINAL);
	  gm.rdma.tail_dma_descriptor = &gm.dma.fake_terminal_dma_descriptor;
	  NOTICE (RDMA_INT_BIT);
	  GM_DECR_DEBUG_CNT (gm.outstanding_dma_cnt);

	  /* We detected 2 DMAs off a single DMA_INT_BIT, so we need
	     to ensure that DMA_INT_BIT was not re-set after
	     we cleared it above.  Since we saw that all terminal
	     bits were set, however, it is safe to assume that all
	     DMA_INT_BIT settings for scheduled DMAs have occurred, so
	     we just need to clear the DMA_INT_BIT. */

	  if (GM_DEBUG_CHAINED_DMA)
	    {
	      if (get_ISR () & DMA_INT_BIT)
		{
		  LOG_DISPATCH (0, "DMA_INT_BIT is set");
		}
	      else
		{
		  LOG_DISPATCH (0, "DMA_INT_BIT is clear");
		}
	      
	    }
	  GM_DEBUG_CHAINED_DMA_ASSERT (gm_chained_dma_slow_all_done ());

	  /* For RDMAs, the wake bit can be arbitrarily delayed until the
	     host accepts the all data from the PCIDMA chip's FIFO.
	     Therefore, we read from the PCI bus to ensure that any
	     WAKE corresponding to the DMAs has completed. */

	  synchronous_RDMA (&gm.trash, gm.page_hash.bogus_sdma_ptr, 4);
	  
	  gm_clear_DMA_INT_BIT ();
	  /* GM_DEBUG_CHAINED_DMA_ASSERT ((get_ISR () & DMA_INT_BIT) == 0); */
	  DISPATCH (4, "Handled DMA_INT_BIT for SDMA and RDMA");
	}
      else
	{
	  /* Only the SDMA completed. */
	  DISPATCH (5, "Handled DMA_INT_BIT for SDMA");
	}
    }
  else if (rdma__next_with_flags & GM_DMA_DESCRIPTOR_TERMINAL)
    {
      /* Only the RDMA completed. */
      
      gm_assert
	(gm.dma.fake_terminal_dma_descriptor.next_with_flags
	 == (gm_lp_n_t) GM_DMA_DESCRIPTOR_TERMINAL);
      gm.rdma.tail_dma_descriptor = &gm.dma.fake_terminal_dma_descriptor;
      NOTICE (RDMA_INT_BIT);
      GM_DECR_DEBUG_CNT (gm.outstanding_dma_cnt);
      DISPATCH (6, "Handled DMA_INT_BIT for RDMA");
    }

  gm_always_assert (!"Bogus DMA_INT_BIT while DMAing.");
}
