/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_galvantech_workaround_h_
#define _gm_galvantech_workaround_h_

/****************************************************************
 * 
 ****************************************************************/

#if GM_ENABLE_GALVANTECH_WORKAROUND

#if 0
/* compute 32-bit one's complement checksum for word-aligned buffer.
   Assumes that it is OK to read 3 word past _limit, and that _start
   and _limit are word aligned. */

gm_u32_t
gm_galvantech_checksum (const void *_start, const void *_limit)
{
  gm_u32_t checksum = 0, a, b, c;
  const gm_u32_t *start, *limit;

  /* Macros to compute ones-complement checksums follow.  These may
     use the carry flag only internally, as the compiler might clobber
     it between macros. */

#define ADDC(c,a) do {							\
  asm ("add.f %1,%2,%0\n\t"						\
       "addc %0,%?r0,%0" : "=r" (c) : "r" (c), "r" (a) : "cc");		\
} while (0)

#define ADDC_3(checksum, start, a, b, c) do {				\
  asm ("add.f %0,%2,%0\n\t"						\
       "ld [%1++],%2\n\t"						\
       "addc.f %0,%3,%0\n\t"						\
       "ld [%1++],%3\n\t"						\
       "addc.f %0,%4,%0\n\t"						\
       "ld [%1++],%4\n\t"						\
       "addc %0,%?r0,%0\n\t"						\
       : "=r"(checksum), "=r"(start), "=r"(a), "=r"(b), "=r"(c)		\
       : "0" (checksum), "1" (start), "2" (a), "3" (b), "4" (c)		\
       : "cc");								\
} while (0)

#define ADDC_15(checksum, start, a, b, c) do {				\
  asm ("add.f %0,%2,%0\n\t"						\
       "ld [%1++],%2\n\t"						\
       "addc.f %0,%3,%0\n\t"						\
       "ld [%1++],%3\n\t"						\
       "addc.f %0,%4,%0\n\t"						\
       "ld [%1++],%4\n\t"						\
       "addc.f %0,%2,%0\n\t"						\
       "ld [%1++],%2\n\t"						\
       "addc.f %0,%3,%0\n\t"						\
       "ld [%1++],%3\n\t"						\
       "addc.f %0,%4,%0\n\t"						\
       "ld [%1++],%4\n\t"						\
       "addc.f %0,%2,%0\n\t"						\
       "ld [%1++],%2\n\t"						\
       "addc.f %0,%3,%0\n\t"						\
       "ld [%1++],%3\n\t"						\
       "addc.f %0,%4,%0\n\t"						\
       "ld [%1++],%4\n\t"						\
       "addc.f %0,%2,%0\n\t"						\
       "ld [%1++],%2\n\t"						\
       "addc.f %0,%3,%0\n\t"						\
       "ld [%1++],%3\n\t"						\
       "addc.f %0,%4,%0\n\t"						\
       "ld [%1++],%4\n\t"						\
       "addc.f %0,%2,%0\n\t"						\
       "ld [%1++],%2\n\t"						\
       "addc.f %0,%3,%0\n\t"						\
       "ld [%1++],%3\n\t"						\
       "addc.f %0,%4,%0\n\t"						\
       "ld [%1++],%4\n\t"						\
       "addc %0,%?r0,%0\n\t"						\
       : "=r"(checksum), "=r"(start), "=r"(a), "=r"(b), "=r"(c)		\
       : "0" (checksum), "1" (start), "2" (a), "3" (b), "4" (c)		\
       : "cc");								\
} while (0)

  if (GM_DEBUG_GALVANTECH_WORKAROUND)
    {
      gm_hex_dump (_start, _limit - _start);
    }

  /* handle misaligned start bytes */

  start = (gm_u32_t *) _start;	/* speculative */
  if ((long) _start & 3)
    {
      switch ((long) _start & 3)
	{
	case 1:
	  start = (gm_u32_t *) (_start + 3);
	  ADDC (checksum, start[-1] & 0x00ffffff);
	  break;
	case 2:
	  start = (gm_u32_t *) (_start + 2);
	  ADDC (checksum, start[-1] & 0x0000ffff);
	  break;
	case 3:
	  start = (gm_u32_t *) (_start + 1);
	  ADDC (checksum, start[-1] & 0x000000ff);
	  break;
	}
    }

  /* prefetch to allow perfect pipelining. */

  a = *start++;
  b = *start++;
  c = *start++;

  /* compute limit of word copies */

  limit = (const gm_u32_t *) GM_ALIGN (lp, _limit, 4);

  /* precompensate for prefetching for start/limit comparisons */

  limit += 3;

  /* handle the bulk of the aligned core with an unrolled loop. */

  while (start < limit - 15)
    {
      ADDC_15 (checksum, start, a, b, c);
    }

  /* handle remaining bulk in smaller unrolled loop */

  while (start < limit - 3)
    {
      ADDC_3 (checksum, start, a, b, c);
    }

  /* handle remaining words  */

  if (start++ < limit)
    {
      ADDC (checksum, a);
      if (start++ < limit)
	{
	  ADDC (checksum, b);
	  if (start < limit)
	    {
	      ADDC (checksum, c);
	    }
	}
    }

  /* handle remaining bytes */

  if ((long) _limit & 3)
    {
      switch ((long) _limit & 3)
	{
	case 3:
	  ADDC (checksum, limit[-3] & 0xffffff00);
	  break;
	case 2:
	  ADDC (checksum, limit[-3] & 0xffff0000);
	  break;
	case 1:
	  ADDC (checksum, limit[-3] & 0xff000000);
	  break;
	}
    }

  /* Rotate computed checksum into the correct position. */

  if ((long) _start & 3)
    {
      switch ((long) _start & 3)
	{
	case 3:
	  checksum = (checksum << 24) | ((checksum >> 8) & 0xffffff);
	  break;
	case 2:
	  checksum = (checksum << 16) | ((checksum >> 16) & 0xffff);
	  break;
	case 1:
	  checksum = (checksum << 8) | ((checksum >> 24) & 0xff);
	  break;
	}
    }

  return checksum;

}

#endif /* 0 */

static inline gm_u32_t
gm_galvantech_header_checksum (const gm_packet_header_t * p)
{
  gm_u32_t *iptr = (gm_u32_t *) p;
  register gm_u32_t checksum;
  register gm_u32_t r0, r1, r2, r3;

  if (((gm_u32_t) p) & 0x3)
    {
      gm_printf ("gm_galvantech_header_checksum(%p) bad pointer\n", p);
      fflush (stdout);
    }
  iptr = (gm_u32_t *) p;
  r0 = iptr[0];
  r1 = iptr[1];
  r2 = iptr[2];
  r3 = iptr[3];
  checksum = r0 + r1;
  checksum = checksum + r2 + r3;
  return (checksum);
}

#endif /* GM_ENABLE_GALVANTECH_WORKAROUND */

static inline void
gm_galvantech_set_header_checksum (gm_packet_header_t * p)
{
#if GM_ENABLE_GALVANTECH_WORKAROUND
  p->header_checksum = gm_galvantech_header_checksum (p);
#else
  GM_PARAMETER_MAY_BE_UNUSED (p);
#endif
}

static inline gm_status_t
gm_galvantech_check_header_checksum (const gm_packet_header_t * p)
{
#if GM_ENABLE_GALVANTECH_WORKAROUND
  gm_u32_t computed;

  computed = gm_galvantech_header_checksum (p);
  if (p->header_checksum != computed)
    {
      gm_u32_t *iptr = (gm_u32_t *) p;
      gm_printf ("*** Bad GM hdr cksum detected 0x%08x!=0x%08x "
		 "(0x%08x) %d->%d.\n",
		 p->header_checksum, computed, (p->header_checksum ^ computed),
		 (int) p->sender_node_id, (int) p->target_node_id);
      fflush (stdout);
      gm_printf ("header: 0x%08x 0x%08x 0x%08x 0x%08x (0x%08x)\n",
		 iptr[0], iptr[1], iptr[2], iptr[3], iptr[4]);
      fflush (stdout);
      gm_printf
	("*** There is probably buggy Galvantech SRAM in node %d or %d\n",
	 (int) p->sender_node_id, (int) p->target_node_id);
      fflush (stdout);
      return GM_FAILURE;
    }
#else
  GM_PARAMETER_MAY_BE_UNUSED (p);
#endif
  return GM_SUCCESS;
}

#if GM_ENABLE_GALVANTECH_WORKAROUND
#define GM_GALVANTECH_SET_IP_CHECKSUM(p, checksum) do {			\
  (p)->ip_checksum = (checksum);					\
} while (0)
#else
#define GM_GALVANTECH_SET_IP_CHECKSUM(p, checksum)
#endif

#endif /* _gm_galvantech_workaround_h_ */
