/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_host_side_tables_h_
#define _gm_host_side_tables_h_

#include "gm_ether_addr.h"

/****************************************************************
 * ethernet address table functions.
 ****************************************************************
 These functions are used to modify entries in the
 host-memory-resident ethernet address table.  This table directly
 informs the host of GM_ID->ethernet_id mappings, and indirectly
 informs the host of the reverse mappings. */

/* In the host-memory-resident ethernet address table, set the
   ethernet address for GM node NODE_ID to ADDR. */

static inline void
gm_ethernet_addr_table_set (unsigned int node_id, gm_u8_t * addr)
{
  gm_dp_t entry_logical_addr;
  gm_dp_t piece;

  /* compute logical address of addr table entry, aborting if not possible. */
  piece = gm.ethernet.addr_table_piece[node_id / (GM_PAGE_LEN / 8)];

  /* Catch case where MCP supports ethernet, but host does not */
  if (!piece)
    return;

  entry_logical_addr = piece + 8 * (node_id % (GM_PAGE_LEN / 8));

  /* Copy the ethernet address to a known 8-byte aligned buffer with
     the last two bytes cleared. */
  ether_addr_copy (addr, gm.ethernet.addr_stage);
  gm.ethernet.addr_stage[6] = gm.ethernet.addr_stage[7] = 0;

  /* DMA the address to the host. */
  synchronous_RDMA (&gm.ethernet.addr_stage, entry_logical_addr, 8);
}

#ifdef DELETE
/* Clear all entries in the host-memory-resident ethernet address table. */

static inline void
gm_ethernet_addr_table_clear (void)
{
  unsigned node_id;

  /* Clear the ethernet address DMA stage */
  gm_bzero (gm.ethernet.addr_stage, 8);

  /* Clear each ethernet address in host memory. */
  for (node_id = 0; node_id <= gm.max_node_id; node_id++)
    gm_ethernet_addr_table_set (node_id, &gm.ethernet.addr_stage[0]);
}
#endif /* DELETE */

/* Insert the host name in the host-resident name table at position NODE_ID */

static inline void
gm_name_table_set (unsigned node_id, gm_u8_t * name)
{
  gm_dp_t piece;
  gm_dp_t entry_logical_addr;

  /* compute the logical address to DMA the name to */
  piece = gm.name_table_piece[node_id / (GM_PAGE_LEN / GM_MAX_HOST_NAME_LEN)];
  if (!piece)
    return;
  entry_logical_addr
    = piece + (GM_MAX_HOST_NAME_LEN
	       * (node_id % (GM_PAGE_LEN / GM_MAX_HOST_NAME_LEN)));

  synchronous_RDMA (name, entry_logical_addr, GM_MAX_HOST_NAME_LEN);
}

#ifdef DELETE
/* Clear the host_resident name table */

static inline void
gm_name_table_clear (void)
{
  char _null_name[GM_MAX_HOST_NAME_LEN + GM_DMA_GRANULARITY];
  char *null_name = GM_DMA_ROUNDUP (lp, _null_name);
  unsigned node_id;

  gm_bzero (_null_name, sizeof (_null_name));
  for (node_id = 0; node_id <= gm.max_node_id; node_id++)
    gm_name_table_set (node_id, null_name);
}
#endif /* DELETE */

#endif /* _gm_host_side_tables_h_ */
