/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_interrupt_h_
#define _gm_interrupt_h_

#include "gm.h"
#include "gm_bootstrap.h"
#include "gm_enable_paranoid_interrupts.h"
#include "gm_flash.h"
#include "gm_lanai_def.h"
#include "gm_lanai_globals.h"
#include "gm_stbar.h"
#include "gmcp.h"

extern void _gm_interrupt (gm_u32_t type);
extern void _gm_interrupt_error (char *error);
extern int _prepared_to_interrupt;

/****************
 * interrupts
 ****************/

static inline
void
_await_low_host_sig_bit (char *LED_msg)
{
  while (get_ISR () & HOST_SIG_BIT)
    {
      gm.while_waiting++;
      gm_flash_while_waiting (LED_msg);
    }
  GM_STBAR ();
}

static inline void
_await_interrupt_completion (char *LED_msg)
{
  _await_low_host_sig_bit (LED_msg);
  GM_STBAR ();
  if (GM_ENABLE_PARANOID_INTERRUPTS && gm.interrupt.type != GM_NO_INTERRUPT)
    {
      unsigned int start;
      /* This happen if writes from host are somehow reordered, be
	 sure we notice and abort, the fix if it happens is generally
	 to renforce the GM_STBAR() macros or to change the attribute
	 used to map IOPAGES */
      start = RTC;
      /* Give the clear interrupt.type transaction some time
	 to land, otherwise it might overwrite our new setting
	 below */
      while ((unsigned)RTC - start < 50000 * 2 
	     && gm.interrupt.type != GM_NO_INTERRUPT)
	{
	  _await_low_host_sig_bit (LED_msg);
	}
      gm.interrupt.failed_assertion.file = __GM_FILE__;
      gm.interrupt.failed_assertion.line = __LINE__;
      gm.interrupt.failed_assertion.text = 
	"interrupt.type was != 0 with ISR_HOST_BIT == 0";
      _gm_interrupt (GM_FAILED_ASSERTION_INTERRUPT);
      while (1)
	gm_morse_sync ("SOS\n");
    }
}

/*
 * Interrupt entry points: Call prepare_to_interrupt() before touching
 * the interrupt structure to ensure that the preceeding interrupt has
 * been handled, then gm_interrupt when done writing it to generate
 * the interrupt.  Finally, you may call await_interrupt_completion
 */

static inline void
prepare_to_interrupt (char *LED_msg)
{
  if (GM_DEBUG && _prepared_to_interrupt)
    _gm_interrupt_error (GM_STR
			 ("recursive interrupt preparation detected."));
  _await_interrupt_completion (LED_msg);
  if (GM_DEBUG)
    _prepared_to_interrupt = 1;
}

static void
gm_interrupt (gm_u32_t type)
{
  if (GM_DEBUG && !_prepared_to_interrupt)
    _gm_interrupt_error (GM_STR ("unprepared interrupt"));
  _gm_interrupt (type);
  _prepared_to_interrupt = 0;
}

static inline void
await_interrupt_completion (char *LED_msg)
{
  if (GM_DEBUG && _prepared_to_interrupt)
    {
      _gm_interrupt_error (GM_STR
			   ("awaiting interrupt while prepared for"
			    " interrupt\n"));
    }
  _await_interrupt_completion (LED_msg);
}

#endif _gm_interrupt_h_
