/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* author: finucane@myri.com */

#define GM_DEBUG_MAPPEE 0

#include "gm_bootstrap.h"
#include "gm_call_trace.h"
#include "gm_ether_addr.h"
#include "gm_mappee.h"
#include "gm_mcp_reboot_node.h"
#include "gm_send_tokens.h"

#if GM_DEBUG_MAPPEE
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/*********************************************************************/
/*this function replies to a scout message sent by the myrinet mapper
  mcp.  scout is the scout message, address is the node address of
  this lanai, mapVersion and mapAddress are version numbers of this
  node's current route table.  If there is no route table, these
  should be zero.  Returns zero if there was an error.*/
/*********************************************************************/	

/* Reply to a mapper scout or config packet. */
void
queue_mapper_scout_reply (struct gm_mapper_scout_reply_stage *mrs,
			  gm_mapper_packet_t * mp)
{
  gm_u8_t *from, *to;
  gm_u32_t route_len;

  gm_assert (mp->common.subtype == GM_MAPPER_SCOUT_PACKET_SUBTYPE);

  mrs->route_len = route_len = mp->scout.route_length;
  if (mrs->route_len <= 32)
    from = &mp->scout.route[mp->scout.route_length];
  else
    from = &mp->scout.extended_route[mp->scout.route_length];

  to = &mrs->route[GM_MAX_NETWORK_DIAMETER];
  if (route_len != GM_NO_ROUTE)
    {
      if (route_len>GM_MAX_NETWORK_DIAMETER) 
	{
	  printf("scout_reply route too long = %d bytes\n",route_len);
	  fflush(stdout);
	  route_len = 0;
	}

      while (route_len--)
	*--to = *--from;
    }

  /* build the packet */
  gm_assert (mrs->packet.type == GM_MAPPING_PACKET_TYPE);
  gm_assert (mrs->packet.subtype == GM_MAPPER_SCOUT_REPLY_PACKET_SUBTYPE);
  mrs->packet.port = mp->scout.port;
  mrs->packet.phase = mp->scout.phase;
  mrs->packet.mapper_open = gm.port[GM_MAPPER_PORT_ID].open;
  /* mrs->packet.address preset */
  /* mrs->packet.gm_id preset */
  /* mrs->packet.map_version preset */

  gm_assert (mrs->packet.option
	     == (GM_MAPPER_SCOUT_REPLY_PACKED_ROUTES_OPTION
		 | GM_MAPPER_SCOUT_REPLY_LONG_HOSTNAME_OPTION));

  /* Schedule the packet to be sent */
  if (!mrs->in_send_queue)
    {
      /* HACK: Be sneaky and put the token in the send queue for port
         0 or target host ID 0, which is guaranteed not to have
         anything in it and therefore is guaranteed not to block. */
      gm_assert (mrs->send_token.common.next == 0);
      append_send_token_to_send_queue (&mrs->send_token, 0, 0);
      mrs->in_send_queue = 1;
      NOTICE (SDMA_PENDING);
    }
}

/* Reply to a mapper scout or config packet. */
void
queue_mapper_config_reply (struct gm_mapper_config_reply_stage *mrs,
			   gm_mapper_packet_t * mp)
{
  gm_u8_t *from, *to;
  gm_u32_t route_len;

  gm_assert (mp->common.subtype == GM_MAPPER_CONFIG_PACKET_SUBTYPE);
  mrs->route_len = route_len = mp->config.bytes[0];
  from = &mp->config.bytes[1 + route_len];

  to = &mrs->route[GM_MAX_NETWORK_DIAMETER];
  if (route_len != GM_NO_ROUTE)
    {
      if (route_len>GM_MAX_NETWORK_DIAMETER) 
	{
	  printf("config_reply route too long = %d bytes\n",route_len);
	  fflush(stdout);
	  route_len = 0;
	}

      while (route_len--)
	*--to = *--from;
    }

  /* build the packet */
  gm_assert (mrs->packet.type == GM_MAPPING_PACKET_TYPE);
  gm_assert (mrs->packet.subtype == GM_MAPPER_CONFIG_REPLY_PACKET_SUBTYPE);
  mrs->packet.port = mp->config.port;
  mrs->packet.phase = mp->config.phase;
  mrs->packet.host_section = mp->config.host_section;
  /* mrs->packet.address preset */

  /* Schedule the packet to be sent */
  if (!mrs->in_send_queue)
    {
      /* HACK: Be sneaky and put the token in the send queue for port
         0 or target host ID 0, which is guaranteed not to have
         anything in it and therefore is guaranteed not to block. */
      gm_assert (mrs->send_token.common.next == 0);
      append_send_token_to_send_queue (&mrs->send_token, 0, 0);
      mrs->in_send_queue = 1;
      NOTICE (SDMA_PENDING);
    }

}
