/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/* Myrinet packet descriptors describe myrinet packets and are used
   to communicate the the gm_send.c and gm_recv.c packet interface state
   engines.  Except for the following, their semantics are pretty
   straightforward.

   gm.send.pending:
   	Packets in this list are use in the obvious fashion.
   gm.send.free:
   gm.recv.free:
        segment[0].ptr points to a send buffer of GM_MYRINET_MTU bytes
	and preceeded by at least GM_MAX_NETWORK_DIAMETER bytes
	reserved for the route.  route.limit = segment[0].ptr.
   gm.recv.pending:
        route.ptr is undefined. route.limit = segment[0].ptr.
	segment[0].ptr points to received packet, and is preceeded by
	at least GM_MAX_NETWORK_DIAMETER bytes reserved for the route
	(to allow forwarding).
o*/

#ifndef _gm_myrinet_packet_descriptor_h_
#define _gm_myrinet_packet_descriptor_h_

#include "gm_list.h"
#include "gm_types.h"
#include "gm_debug.h"
#include "gm_dispatch.h"

#define GM_DEBUG_MPDS 0

/****************************************************************
 * Prototypes
 ****************************************************************/

extern void _gm_mpd_print (gm_myrinet_packet_descriptor_t *d);
gm_status_t gm_mpd_create (gm_myrinet_packet_descriptor_free_list_t *fl,
			   unsigned long mtu);
void gm_mpd_free_list_finalize (gm_myrinet_packet_descriptor_free_list_t *fl);
gm_status_t gm_mpd_free_list_init (gm_myrinet_packet_descriptor_free_list_t
				   *fl,
				   unsigned cnt,
				   unsigned mtu,
				   gm_u32_t nonempty_state_bit);
extern gm_status_t gm_mpd_init (void);
void gm_mpd_debug_report (void);

/****************************************************************
 * Inlines
 ****************************************************************/

#define gm_mpd_print(d) do {						\
  gm_myrinet_packet_descriptor_t *gm_mpd_print__d;			\
									\
  gm_mpd_print__d = (d);						\
  gm_assert (gm_mpd_print__d);						\
  _gm_mpd_print (gm_mpd_print__d);					\
} while (0)

static inline
gm_myrinet_packet_descriptor_t *
gm_mpd_alloc (gm_myrinet_packet_descriptor_free_list_t *free_list)
{
  gm_myrinet_packet_descriptor_t *ret;
  gm_myrinet_packet_descriptor_t *ret__next;
  gm_u32_t nonempty_state_bit;

  GM_CALLED_WITH_ARGS (("%p", free_list));
  
  ret = free_list->first_free;
  gm_assert (ret);
  nonempty_state_bit = free_list->nonempty_state_bit;
  ;
  ret__next = ret->next;
  ret->next = 0;
  free_list->first_free = ret__next;
  if (!ret__next)
    {
      gm_assert (NOTICED (nonempty_state_bit));
      TOGGLE_STATE (nonempty_state_bit);
    }

  gm_assert (ret->route.limit == ret->header.ptr
	     || free_list != &gm.send.free_list);
  
  GM_RETURN_PTR (ret);
}

static inline
gm_myrinet_packet_descriptor_t *
gm_mpd_maybe_alloc (gm_myrinet_packet_descriptor_free_list_t *free_list)
{
  gm_myrinet_packet_descriptor_t *ret;
  gm_myrinet_packet_descriptor_t *ret__next;
  gm_u32_t nonempty_state_bit;

  GM_CALLED_WITH_ARGS (("%p", free_list));
  
  ret = free_list->first_free;
  nonempty_state_bit = free_list->nonempty_state_bit;
  ;
  ret__next = ret->next;	/* HACK: OK to dereference NULL on LANai */
  if (!ret)
    {
      GM_RETURN_PTR (ret);
    }
  ret->next = 0;
  free_list->first_free = ret__next;
  if (!ret__next)
    {
      gm_assert (NOTICED (nonempty_state_bit));
      TOGGLE_STATE (nonempty_state_bit);
    }

  GM_RETURN_PTR (ret);
}

/* append descriptor to list */

static inline
void
_gm_mpd_list_append (gm_myrinet_packet_descriptor_t *d,
		    gm_myrinet_packet_descriptor_list_t *list)
{
  gm_myrinet_packet_descriptor_t *list__last;
  gm_u32_t nonempty_state_bit;
  
  GM_CALLED_WITH_ARGS (("%p, %p", d, list));
  
  nonempty_state_bit = list->nonempty_state_bit;
  list__last = list->last;
  list->last = d;
  gm_assert (d->next == 0);
  NOTICE (nonempty_state_bit);
  list__last->next = d;

  GM_RETURN_NOTHING ();
}

#define gm_mpd_list_append(_d, _list) do {				\
  gm_myrinet_packet_descriptor_t *gm_mpd_list_append_d;			\
  gm_myrinet_packet_descriptor_list_t *gm_mpd_list_append_list;		\
									\
  gm_mpd_list_append_d = (_d);						\
  gm_mpd_list_append_list = (_list);					\
  gm_assert (gm_mpd_list_append_d);					\
  gm_assert (gm_mpd_list_append_list);					\
  gm_assert (gm_mpd_list_append_d->next == 0);				\
  GM_ASSERT_NOT_IN_LIST (gm_myrinet_packet_descriptor_t,		\
			 gm_mpd_list_append_list->first,		\
			 gm_mpd_list_append_d, next);			\
  _gm_mpd_list_append (gm_mpd_list_append_d, gm_mpd_list_append_list);	\
} while (0)


#if 0
/* prepend descriptor to list.  This is usually a bad idea, since it
   does not maintain message order, but it is useful for an
   optimization that prepends (N)ACK packets to the rdma list in a
   recv handler and then immediately invokes the rdma handler as a
   sort of subroutine call. */

static inline
void
gm_mpd_list_prepend (gm_myrinet_packet_descriptor_t *d,
		    gm_myrinet_packet_descriptor_list_t *list)
{
  gm_myrinet_packet_descriptor_t *list__first;
  gm_u32_t nonempty_state_bit;
  
  GM_CALLED_WITH_ARGS (("%p, %p", d, list));
  
  list__first = list->first;
  nonempty_state_bit = list->nonempty_state_bit;
  list->first = d;
  d->next = list__first;
  NOTICE (nonempty_state_bit);
  if (!list__first)
    {
      list->last = d;
    }
  
  GM_RETURN_NOTHING ();
}
#endif

static inline
void
gm_mpd_list_init (gm_myrinet_packet_descriptor_list_t *list,
		  gm_u32_t nonempty_state_bit)
{
  GM_CALLED_WITH_ARGS (("%p, 0x%x", list, nonempty_state_bit));
  
  gm_bzero (list, sizeof (*list));
  list->nonempty_state_bit = nonempty_state_bit;

  GM_RETURN_NOTHING ();
}

#define gm_mpd_free(_d) do {						\
  gm_myrinet_packet_descriptor_t *gm_mpd_free_d;			\
									\
  gm_mpd_free_d = (_d);							\
  gm_assert (gm_mpd_free_d);						\
  gm_assert (gm_mpd_free_d->free_list);					\
  gm_assert (gm_mpd_free_d != gm_mpd_free_d->free_list->first_free);	\
  GM_ASSERT_NOT_IN_LIST (gm_myrinet_packet_descriptor_t,		\
			 gm_mpd_free_d->free_list->first_free,		\
			 gm_mpd_free_d, next);				\
  _gm_mpd_free (gm_mpd_free_d);						\
} while (0)

static inline
void
_gm_mpd_free (gm_myrinet_packet_descriptor_t *d)
{
  gm_myrinet_packet_descriptor_callback_lp_t d__free_callback;
  gm_myrinet_packet_descriptor_free_list_t *free_list;
  gm_myrinet_packet_descriptor_t *d__next;
  gm_myrinet_packet_descriptor_t *free_list__first_free;
  gm_u32_t free_list__nonempty_state_bit;
  
  GM_CALLED_WITH_ARGS (("%p", d));

  gm_assert (d);
  gm_assert (d->route.limit == d->header.ptr
	     || d->free_list != &gm.send.free_list);
  
  free_list = d->free_list;
  d__next = d->next;
  d->next = 0;
  free_list__nonempty_state_bit = free_list->nonempty_state_bit;
  d__free_callback = d->free_callback;
  free_list__first_free = free_list->first_free;
  gm_assert (free_list__first_free != d);
  free_list->first_free = d;
  NOTICE (free_list__nonempty_state_bit);
  d->next = free_list__first_free;
  if (d__free_callback)
    {
      void *d__free_callback_context;
      
      d__free_callback_context = d->free_callback_context;
      d__free_callback (d, d__free_callback_context);
    }

  if (GM_DEBUG)
    {
      free_list->debug_free_cnt++;
    }
  
  GM_RETURN_NOTHING ();
}

#define gm_mpd_list_free_head(_d, _list) do {				\
  gm_myrinet_packet_descriptor_t *gm_mpd_d;				\
  gm_myrinet_packet_descriptor_list_t *gm_mpd_list;			\
									\
  gm_mpd_d = (_d);							\
  gm_mpd_list = (_list);						\
  gm_assert (gm_mpd_list->first == gm_mpd_d);				\
  gm_assert (gm_mpd_d->free_list);					\
  GM_ASSERT_NO_CYCLE (gm_myrinet_packet_descriptor_t,			\
		      gm_mpd_d, next);					\
  GM_ASSERT_NOT_IN_LIST (gm_myrinet_packet_descriptor_t,		\
			 gm_mpd_d->free_list->first_free, gm_mpd_d,	\
			 next);						\
  _gm_mpd_list_free_head (gm_mpd_d, gm_mpd_list);			\
} while (0)

static inline
void
_gm_mpd_list_free_head (gm_myrinet_packet_descriptor_t *d,
		       gm_myrinet_packet_descriptor_list_t *list)
{
  gm_myrinet_packet_descriptor_callback_lp_t d__free_callback;
  gm_myrinet_packet_descriptor_free_list_t *free_list;
  gm_myrinet_packet_descriptor_t *d__next;
  gm_myrinet_packet_descriptor_t *free_list__first_free;
  gm_u32_t free_list__notice_on_free;
  gm_u32_t toggle_on_list_empty;

  GM_CALLED_WITH_ARGS (("%p, %p", d, list));
  
  gm_assert (list->first == d);
  gm_assert (NOTICED (list->nonempty_state_bit));
  gm_assert (d->next != d);	/* check for simple cycle */
  
  free_list = d->free_list;
  d__next = d->next;
  toggle_on_list_empty = list->nonempty_state_bit;
  free_list__notice_on_free = free_list->nonempty_state_bit;
  free_list__first_free = free_list->first_free;
  free_list->first_free = d;
  list->first = d__next;
  d->next = free_list__first_free;
  NOTICE (free_list__notice_on_free);
  gm_assert (NOTICED (toggle_on_list_empty));
  d__free_callback = d->free_callback;
  if (!d__next)
    {
      list->last
	= ((gm_myrinet_packet_descriptor_t *)
	   ((char *) &list->first
	    - GM_OFFSETOF (gm_myrinet_packet_descriptor_t, next)));
      TOGGLE_STATE (toggle_on_list_empty);
      LOG_DISPATCH (20, "list is empty");
    }
  else
    {
      LOG_DISPATCH (21, "list is not empty");
    }
  if (d__free_callback)
    {
      void *d__free_callback_context;

      LOG_DISPATCH (20, "calling descriptor free_callback");
      d__free_callback_context = d->free_callback_context;
      ;
      ;
      d__free_callback (d__free_callback_context, d);
      LOG_DISPATCH (20, "called descriptor free_callback");
    }
  else
    {
      LOG_DISPATCH (20, "no descriptor free_callback");
    }
  
  GM_RETURN_NOTHING ();
}

#define gm_mpd_list_remove_head(_d, _list) do {				\
  gm_myrinet_packet_descriptor_t *gm_mpd_d;				\
  gm_myrinet_packet_descriptor_list_t *gm_mpd_list;			\
									\
  gm_mpd_d = (_d);							\
  gm_mpd_list = (_list);						\
  gm_assert (gm_mpd_d->next != gm_mpd_d);				\
  _gm_mpd_list_remove_head (gm_mpd_d, gm_mpd_list);			\
} while (0)
     
static inline
void
_gm_mpd_list_remove_head (gm_myrinet_packet_descriptor_t *d,
			  gm_myrinet_packet_descriptor_list_t *list)
{
  gm_myrinet_packet_descriptor_t *d__next;
  gm_u32_t toggle_on_list_empty;

  GM_CALLED_WITH_ARGS (("%p, %p", d, list));
  
  gm_assert (list->first == d);
  gm_assert (NOTICED (list->nonempty_state_bit));
  
  d__next = d->next;
  toggle_on_list_empty = list->nonempty_state_bit;
  d->next = 0;
  list->first = d__next;
  gm_assert (NOTICED (toggle_on_list_empty));
  if (!d__next)
    {
      list->last
	= ((gm_myrinet_packet_descriptor_t *)
	   ((char *) &list->first
	    - GM_OFFSETOF (gm_myrinet_packet_descriptor_t, next)));
      TOGGLE_STATE (toggle_on_list_empty);
      LOG_DISPATCH (22, "list is empty");
    }
  else
    {
      LOG_DISPATCH (23, "list is not empty");
    }
  

  GM_RETURN_NOTHING ();
}

/* start a receive into the buffer described by d. */

static inline
void
gm_mpd_recv_into (gm_myrinet_packet_descriptor_t *d)
{
  void *start;

  gm_assert (d);
  
  start = d->header.ptr;
  gm.recv.descriptor_being_filled = d;
  RMP = start;
  set_RML (start + GM_MYRINET_MTU + 8);
  NOTICE (RECEIVING);
}

#endif /* _gm_myrinet_packet_descriptor_h_ */
