/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_alarm.h"
#include "gm_connection.h"
#include "gm_lanai_globals.h"
#include "gm_port.h"
#include "gm_recv_token_hash.h"

void
remove_port_from_polling_queue (gm_port_protected_lanai_side_t * port)
{
  if (!port->next_to_poll)
    return;

  if (port->next_to_poll == port)
    {
      gm.poll.port = 0;		/* mark polling queue as empty */
      gm.handler[POLL_EVENT] = gm.poll.idle_handler;	/* stop polling */
    }
  else
    {
      gm_port_protected_lanai_side_t **p;

      /* arrange to not poll the port we are about to remove */

      if (gm.poll.port == port)
	{
	  gm.poll.port = port->next_to_poll;
	}

      /* remove the port from the circular queue */

      for (p = &port->next_to_poll; *p != port; p = &(*p)->next_to_poll)
	;
      *p = port->next_to_poll;
    }

  port->next_to_poll = 0;	/* mark as not in queue */
}

void
open_port (gm_port_protected_lanai_side_t * port)
{
  unsigned int j;

  GM_PRINT (GM_DEBUG_CONNECTIONS, ("opening port %u\n", port->id));

  gm_assert (!port->open);
  gm_assert (!port->next_to_poll);
  gm_assert (port->active_subport_cnt == 0);

  /* Reset the send list for this port. */

  port->first_free_send_token = &port->_send_tokens[0];
  for (j = 0; j < GM_NUM_ELEM (port->_send_tokens) - 1; j++)
    port->_send_tokens[j].common.next = &port->_send_tokens[j + 1];
  port->last_free_send_token = &port->_send_tokens[j];
  port->last_free_send_token->common.next = 0;
  port->send_token_queue_slot = &port->PORT->send_token_queue[0];

  /* Reset the send list for this port. */

  port->recv_queue_slot_num = 0;

  /* insert the port in the polling queue. */

  if (!gm.poll.port)
    {
      port->next_to_poll = port;
      gm.poll.port = port;
      gm.handler[POLL_EVENT] = gm.poll.active_handler;	/* start polling */
    }
  else
    {
      port->next_to_poll = gm.poll.port->next_to_poll;
      gm.poll.port->next_to_poll = port;
    }
  port->open = -1;
}

void
gm_init_port (unsigned p)
{
  unsigned int j;
  gm_port_protected_lanai_side_t *port;

  GM_PRINT (GM_DEBUG_CONNECTIONS, ("initializing port %u\n", p));

  port = &gm_port[p];

/***** unprotected port initialization *****/

  /* all fields cleared. */

  gm_assert (GM_PORT);
  gm_bzero (&GM_PORT[p], sizeof (gm_port_unprotected_lanai_side_t));

/****** protected port initialization ******/

  gm_bzero (&gm_port[p], sizeof (gm_port_protected_lanai_side_t));

  port->send_token_queue_slot = &GM_PORT[p].send_token_queue[0];
  port->recv_token_queue_slot = &GM_PORT[p].recv_token_queue[0];
  /* port->woke_host = 0; */
  port->first_free_send_token = &port->_send_tokens[0];
  for (j = 0; j < GM_NUM_ELEM (port->_send_tokens) - 1; j++)
    port->_send_tokens[j].common.next = &port->_send_tokens[j + 1];
  port->last_free_send_token = &port->_send_tokens[j];

  port->free_recv_tokens = &port->_recv_tokens[0];
  for (j = 0; j < GM_NUM_ELEM (port->_recv_tokens) - 1; j++)
    port->_recv_tokens[j].next = &port->_recv_tokens[j + 1];
  /* gm.free_recv_tokens(i)[][] cleared */

  /* Initialize the recv queue such that DMAs to it are harmless. */

  for (j = 0; j < GM_NUM_ELEM (port->recv_queue_slot_dma_addr); j++)
    port->recv_queue_slot_dma_addr[j] = gm.page_hash.bogus_rdma_ptr;

  /* port->sent_slot = 0; */
  port->sent.type = GM_NEW_SENT_TOKENS_EVENT;

  port->id = p;
  port->PORT = &GM_PORT[p];
  port->next_to_poll = 0;
  port->open = 0;
}

/* Reset the port such that all messages will be rejected, and all
   sends are aborted. */

#define GM_DEBUG_RESET_PORT 0

void
reset_port (unsigned port_id)
{
  unsigned int i;
  gm_port_protected_lanai_side_t *port;
  unsigned int active_subport_cnt;

  GM_PRINT (GM_DEBUG_CONNECTIONS, ("resetting port %u\n", port_id));
  if (GM_DEBUG_RESET_PORT)
    {
      gm_printf (GM_STR (__FUNCTION__ "() called\n"));
    }

  /* When the mapper port is closed, set the mapper level for this
     node to 0. */

  if (port_id == GM_MAPPER_PORT_ID)
    {
      gm.mapper_state.scout_reply.packet.level = 0;
    }

  port = &gm_port[port_id];
  if (!port->open)
    return;

  /* preserve the active subport count */

  active_subport_cnt = port->active_subport_cnt;
  if (GM_DEBUG_RESET_PORT)
    {
      gm_printf (GM_STR (__FUNCTION__ " active_subport_cnt = %d\n"),
		 active_subport_cnt);
    }

  /************
   * reset unprotected port state
   ************/

  /* Clear any token in the user queues. */

  gm_assert (GM_PORT);
  gm_bzero (&GM_PORT[port_id], sizeof (GM_PORT[port_id]));

  /************
   * Reset the protected port state
   ************/

  if (GM_DEBUG_RESET_PORT)
    {
      gm_printf (GM_STR (__FUNCTION__ " unsetting alarm\n"));
    }

  if (port->alarm_set)
    gm_unset_alarm (port);

  /* mark the port as closed. */

  port->privileged = 0;
  port->wake_host = 0;
  port->enable_nack_down_flag = 0;
  port->open = 0;

  /* Arrange for all sends for the port to be dequeued the next time
     they are acked, nacked, or timed out.  The driver will prevent
     the port from being opened until this is done by checking
     port->active_subport_cnt is zero before allowing the port to be
     opened.

     Find any associated subport that has been disabled, and reenable
     it, allowing the messages to be resent and accepted or dropped. */

  if (GM_DEBUG_RESET_PORT)
    {
      gm_printf (GM_STR (__FUNCTION__ " deactivating connections\n"));
    }

  if (GM_DEBUG)
    check_subport_cnt ();

  if (gm.first_active_connection)
    {
      gm_u32_t c_cnt = 0;
      gm_connection_t *c;

      /* count the open connections */

      c = gm.first_active_connection;
      do
	c_cnt++;
      while ((c = c->next_active) != gm.first_active_connection);

      /* scan connections */

      while (c_cnt--)
	{
	  gm_connection_t *next_c = c->next_active;
	  gm_subport_t *sp = c->first_active_send_port;
	  gm_u32_t sp_cnt = 0;

	  /* count the subports for the connection */

	  do
	    {
	      sp_cnt++;
	      sp = sp->next;
	    }
	  while (sp != c->first_active_send_port);

	  /* scan the subports */

	  gm_assert (sp);
	  while (sp_cnt--)
	    {
	      gm_subport_t *next_sp;

	      next_sp = sp->next;
	      if (GM_SUBPORT_PORT (sp->id) == port_id)
		{
		  gm_assert (active_subport_cnt);

		  GM_PRINT (GM_DEBUG_CONNECTIONS,
			    ("enabling subport %u\n", (int) sp->id));
		  if (sp->disabled)
		    {
		      gm_send_token_t *st;

		      /* If we used a hack to keep the disabled
		         subport around, use another hack to perform
		         the deferred
		         remove_first_send_token_from_send_queue() */

		      if (sp->first_send_token == 0)
			{
			  gm_send_token_t hack;

			  gm_bzero (&hack, sizeof (hack));
			  hack.common.subport = sp;
			  sp->first_send_token = &hack;
			  remove_first_send_token_from_send_queue (&hack,
								   sp->id);
			}

		      sp->disabled = 0;
		      while ((st = sp->first_send_token) != 0)
			{
			  remove_first_send_token_from_send_queue
			    (st, sp->id);
			  pass_sent_token_to_port_and_free
			    (st, &gm.port[GM_SUBPORT_PORT (sp->id)], 0,
			     GM_SEND_PORT_CLOSED);
			}
		      active_subport_cnt--;
		    }
		}
	      sp = next_sp;
	    }
	  c = next_c;
	}
    }
  if (GM_DEBUG_RESET_PORT)
    {
      gm_printf (GM_STR (__FUNCTION__ " done deactivating connections\n"));
    }

  /* defer resetting the send token queues until the port is opened.
     Only clear the queue. */

  for (i = 0; i < GM_NUM_ELEM (port->_send_tokens); i++)
    port->PORT->send_token_queue[i].type = 0;

  /* Remove recv tokens for receives on this port */

  /* Free recv tokens in hash table. */
  recv_token_hash_remove_port_references (port_id);
  /* Free recv tokens in recv queue */

  port->free_recv_tokens = &port->_recv_tokens[0];

  for (i = 0; i < GM_NUM_RECV_TOKENS - 1; i++)
    port->_recv_tokens[i].next = &port->_recv_tokens[i + 1];

  if (GM_DEBUG_RESET_PORT)
    {
      gm_printf (GM_STR (__FUNCTION__ "3\n"));
    }

  for (i = 0; i < GM_NUM_SIZES; i++)
    {
      port->free_recv_token[GM_LOW_PRIORITY][i] = 0;
      port->free_recv_token[GM_HIGH_PRIORITY][i] = 0;
    }

  if (GM_DEBUG_RESET_PORT)
    {
      gm_printf (GM_STR ("clearing recv queue\n"));
    }

  /* Clear the recv queue in host memory. */
  {
    gm_u8_t reserved_space[sizeof (gm_recv_queue_slot_t) +
			   GM_DMA_GRANULARITY];
    gm_recv_queue_slot_t *clear;

    clear = GM_DMA_ROUNDUP (lp, reserved_space);
    gm_bzero (clear, sizeof (*clear));
    for (port->recv_queue_slot_num = 0;
	 port->recv_queue_slot_num < GM_NUM_RECV_QUEUE_SLOTS;
	 port->recv_queue_slot_num++)
      {
	gm_dp_t ear;

	ear = RECV_QUEUE_SLOT_DMA_ADDR (port);
	gm_assert (GM_DMA_ALIGNED (sizeof (*clear)));
	if (ear)		/* catch uninitialized port closure case */
	  {
	    synchronous_GRANULAR_RDMA
	      (clear, ear, (unsigned int) sizeof (*clear));
	  }
      }

    port->recv_queue_slot_num = 0;
    port->recv_token_queue_slot = &GM_PORT[port_id].recv_token_queue[0];
  }
  if (GM_DEBUG_RESET_PORT)
    {
      gm_printf (GM_STR ("cleared\n"));
    }

  /* Initialize the recv queue such that DMAs to it are harmless */

  for (i = 0; i < GM_NUM_RECV_QUEUE_SLOTS; i++)
    port->recv_queue_slot_dma_addr[i] = gm.page_hash.bogus_rdma_ptr;

  if (port_id == GM_ETHERNET_PORT_ID)
    {
      unsigned int j;
      /* this is probably not complete, there may be some other
         variables that need to be reset */
      /* update 1999/07/16: resetting token_slot may be the last one */
      for (i = 0; i < GM_NUM_ETHERNET_RECV_TOKENS; i++)
	for (j = 0; j < GM_MAX_ETHERNET_SCATTER_CNT; j++)
	  gm.ethernet.recv.token[i].segment[j].len = 0;
      gm.ethernet.recv.token_slot = &gm.ethernet.recv.token[0];
      gm.ethernet.recv.remaining_len = 0;
      gm.ethernet.send.gather_cnt = 0;
    }

  /* next_with_sent_packets updated when sends complete
     active_subport_cnt updated when sends complete */

  /* Remove the port from the list of ports with sent messages. */

  if (GM_DEBUG_RESET_PORT)
    {
      gm_printf (GM_STR ("updating sent queues\n"));
    }
  update_host_sent_queues ();
  if (GM_DEBUG_RESET_PORT)
    {
      gm_printf (GM_STR ("updated sent queues\n"));
    }

  /* port->next_with_alarm updated above */
  /* port->alarm_time updated above */
  port->id = port_id;

  port->unacceptable_recv_sizes[GM_LOW_PRIORITY] = 0;
  port->unacceptable_recv_sizes[GM_HIGH_PRIORITY] = 0;

  /* port->PORT unchanged */

  /************
   * Remove remaining global references to the port
   ************/

  if (gm.registered_raw_recv_port == port)
    gm.registered_raw_recv_port = 0;
  remove_port_from_polling_queue (port);

  /* restore the active subport count */

  port->active_subport_cnt = active_subport_cnt;
}

