/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_recv_queue_h_
#define _gm_recv_queue_h_

#include "gm_dispatch.h"
#include "gm_internal_funcs.h"
#include "gm_interrupt.h"

#define GM_DEBUG_RECV_QUEUE 0

static inline void
wake_port (gm_port_protected_lanai_side_t * port)
{
  port->wake_host = 0;

  /* Tell host driver to wake port. */

  prepare_to_interrupt ("333");
  gm.interrupt.wake.port = port->id;
  gm_interrupt (GM_WAKE_INTERRUPT);
}

/* If requested, wake host */
static inline void
wake_port_if_needed (gm_port_protected_lanai_side_t * port)
{
  if (port->wake_host)
    wake_port (port);
}

static inline gm_dp_t
RECV_QUEUE_SLOT_DMA_ADDR (gm_port_protected_lanai_side_t *port)
{
  return port->recv_queue_slot_dma_addr[port->recv_queue_slot_num];
}

static inline gm_up_t
RECV_QUEUE_SLOT_HOST_ADDR (gm_port_protected_lanai_side_t *port)
{
  return port->recv_queue_slot_host_addr[port->recv_queue_slot_num];
}

static inline void
RECV_QUEUE_ADVANCE (gm_port_protected_lanai_side_t *port) {
  if (++port->recv_queue_slot_num >= GM_NUM_RECV_QUEUE_SLOTS)
    port->recv_queue_slot_num = 0;
  GM_PRINT (GM_DEBUG_RECV_QUEUE, ("new recv queue slot for port %d is %d\n",
				  port->id, port->recv_queue_slot_num));
}

/* Place a trivial event in the receive queue in host memory. */

static inline void
gm_report (gm_port_protected_lanai_side_t * port,
	   enum gm_recv_event_type type)
{
  /* Verify user legitimate "NEW" type was specified. */

  gm_assert (GM_POWER_OF_TWO (GM_NEW_NO_RECV_EVENT));
  gm_assert (type & GM_NEW_NO_RECV_EVENT);

  /* Report the event */
  gm.report_dma_stage.type = type;

  gm_assert (sizeof (gm.report_dma_stage) == GM_RDMA_GRANULARITY);
  synchronous_GRANULAR_RDMA (&gm.report_dma_stage,
			     (RECV_QUEUE_SLOT_DMA_ADDR (port)
			      + sizeof (gm_recv_queue_slot_t)
			      - GM_RDMA_GRANULARITY),
			     GM_RDMA_GRANULARITY);

  wake_port_if_needed (port);

  /* Advance the receive queue */

  RECV_QUEUE_ADVANCE (port);
}

/* non-inline version of gm_report() to save space for the
   non-performance critical error reporting task. */

void gm_report_error (gm_port_protected_lanai_side_t * port,
		      enum gm_recv_event_type type);

#endif /* _gm_recv_queue_h_ */
