/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_recv_token_h_
#define _gm_recv_token_h_

#include "gm_recv_queue.h"

extern int recv_token_cnt[GM_NUM_PRIORITIES][GM_NUM_SIZES];

/* Move all recv tokens queued by the host for PRIORITY into the free list
   for the appropriate size and priority. */

/* Move all recv tokens queued by the host for PRIORITY into the free list
   for the appropriate size and priority. */

static inline void
absorb_recv_tokens_from_host (gm_port_protected_lanai_side_t * port)
{
  gm_host_recv_token_t *hrt;
  unsigned size, priority;
  int ready;

  while (1)
    {
      gm_recv_token_t *port__free_recv_tokens;
      gm_port_unprotected_lanai_side_t *PORT;

      /* find a host recv token to absorb, if any */

      hrt = port->recv_token_queue_slot;
      /* prefetch */ port__free_recv_tokens = port->free_recv_tokens;
      /* prefetch */ PORT = port->PORT;
      ready = hrt->ready;
      GM_STBAR ();
      /* prefetch */ size = hrt->size;
      /* prefetch */ priority = hrt->priority;
      if (ready)
	{
	  gm_up_t message;
	  unsigned int tag;


	  /* Read rest of info from recv token */
	  message = hrt->message;
	  tag = hrt->tag;

	  /* Catch user violations of recv token regulation. */

	  if (port__free_recv_tokens)
	    {
	      /* Advance recv token queue.  Do this before checking
	         token fields to prevent multiple bad token
	         reports.

	         HACK: Use a fullword write, also clobbering some other
	         fields, to minimize the use of partword stores. */

	      gm_assert (GM_OFFSETOF (gm_host_recv_token_t, tag) + 3
			 == GM_OFFSETOF (gm_host_recv_token_t, ready));
	      gm_assert (GM_OFFSETOF (gm_host_recv_token_t, priority) + 1
			 == GM_OFFSETOF (gm_host_recv_token_t, size));
	      gm_assert (GM_OFFSETOF (gm_host_recv_token_t, size) + 1
			 == GM_OFFSETOF (gm_host_recv_token_t, ready));
	      *(gm_u32_t *) & hrt->ready = 0;
	      port->recv_token_queue_slot = hrt + 1;

	      /* Drop tokens of illegitimate size. Don't check the
	         message pointer P, as the page hash table will catch
	         errors there. */

	      if (size < GM_NUM_SIZES)
		{
		  if (priority <= GM_MAX_PRIORITY)
		    {
		      gm_recv_token_t *t, *next_free;
		      gm_recv_token_t **ready_list, *first_ready;

		      if (GM_DEBUG_RECV_TOKENS)
			{
			  gm_printf (GM_STR
				     ("0x%x tokens, priority %d size 0x%x\n"),
				     ++recv_token_cnt[priority][size],
				     priority, size);
			  fflush (stdout);
			}

		      /* Allocate a receive token buffer in which to
		         store the recv token info */

		      t = port__free_recv_tokens;
		      next_free = t->next;
		      gm_assert (t);
		      ready_list = &port->free_recv_token[priority][size];
		      first_ready = *ready_list;
		      port->free_recv_tokens = next_free;

		      /* Initialize token, and add it to the appropriate
		         ready list. */

		      t->orig_ptr = message;
		      t->recv_ptr = message;
		      t->tag = tag;
		      t->next = first_ready;
		      *ready_list = t;
		    }
		  else
		    {
#if 0
		      /* We really don't wan't to waste tons of memory
		         inlining this repeatedly.  This belongs in a
		         function in mcp/gm_error_handlers.c */
		      gm_printf ("bad recv token - priority pri=%d"
				 " max_pri=%d\n", priority, GM_MAX_PRIORITY);
/*
                      gm_printf("start = 0x%p  end = 0x%p\n",
				&PORT->recv_token_queue[0],
				(&PORT->recv_token_queue
				[GM_NUM_RECV_TOKEN_QUEUE_SLOTS]));
*/
		      fflush (stdout);
		      gm_hex_dump (((unsigned char *) hrt),
				   sizeof (*hrt) + 32);
		      fflush (stdout);
#endif
		      gm_report_error (port, GM_NEW_BAD_RECV_TOKEN_EVENT);
		      continue;
		    }
		}
	      else
		{
#if 0
		  /* We really don't wan't to waste tons of memory
		     inlining this repeatedly.  This belongs in a
		     function in mcp/gm_error_handlers.c */
		  gm_printf ("bad recv token - size   size=%d max_size=%d\n",
			     size, GM_NUM_SIZES);
		  fflush (stdout);
#endif
		  gm_report_error (port, GM_NEW_BAD_RECV_TOKEN_EVENT);
		  continue;
		}
	    }
	  else
	    {
	      gm_report_error (port, GM_NEW_RECV_TOKEN_VIOLATION_EVENT);
	      return;
	    }
	}
      else
	{
	  if (hrt < &PORT->recv_token_queue[GM_NUM_RECV_TOKEN_QUEUE_SLOTS])
	    return;
	  hrt = port->recv_token_queue_slot = &PORT->recv_token_queue[0];
	  if (!hrt->ready)
	    return;
	}
    }
}

/* Allocate a token for a receive of PRIORITY and SIZE, if
   possible. */

static inline gm_recv_token_t *
alloc_lanai_recv_token (unsigned priority,
			gm_port_protected_lanai_side_t * port, unsigned size)
{
  gm_recv_token_t *first;

  first = port->free_recv_token[priority][size];

  if (!first)
    {
      absorb_recv_tokens_from_host (port);
      first = port->free_recv_token[priority][size];
      if (!first)
	return first;
    }

  port->free_recv_token[priority][size] = first->next;

  if (GM_DEBUG_RECV_TOKENS)
    {
      gm_printf (GM_STR ("0x%x tokens, priority %d size 0x%x\n"),
		 --recv_token_cnt[priority][size], priority, size);
      fflush (stdout);
    }

  return first;
}

#endif /* _gm_recv_token_h_ */
