/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_routes_h_
#define _gm_routes_h_

#include "gm_host_side_tables.h"

static inline void
gm_mark_routes_as_retired (void)
{
  unsigned node_id;
  for (node_id = 0; node_id <= gm.max_node_id; node_id++)
    gm_connection[node_id].retired = 1;
}

/* Clear irrrelevant entries in the host-memory-resident address &
   hostname tables. */
static inline void
gm_clear_retired_routes (void)
{
  unsigned node_id;
  gm_u8_t addr_stage[8];
  char _null_name[GM_MAX_HOST_NAME_LEN + GM_DMA_GRANULARITY];
  char *null_name = GM_DMA_ROUNDUP (lp, _null_name);

  gm_bzero (addr_stage, 8);
  gm_bzero (_null_name, sizeof (_null_name));

  /* Clear each ethernet address in host memory. */
  for (node_id = 0; node_id <= gm.max_node_id; node_id++)
    {
      if (gm_connection[node_id].retired)
	{
	  GM_CONNECTION_CLEAR_ROUTE (&gm_connection[node_id]);
	  gm_connection[node_id].retired = 0;
	  gm_ethernet_addr_table_set (node_id, &addr_stage[0]);
	  gm_name_table_set (node_id, null_name);
	}
    }
}

#if GM_MAX_NETWORK_DIAMETER != 24
#error Code does not support GM_MAX_NETWORK_DIAMETER != 24.
#endif

/* hand-optimized route copying function */

static inline void
copy_route (gm_u8_t * _from, gm_u8_t * _to)
{
  gm_u32_t a, b;
  gm_u32_t *from, *to;

  gm_assert (GM_MAX_NETWORK_DIAMETER == 24);

  gm_assert (GM_ALIGNED (_from, 4));
  from = (gm_u32_t *) _from;

  gm_assert (GM_ALIGNED (_to, 4));
  to = (gm_u32_t *) _to;

#if defined lanai7
  /* optimized for LANai7 pipeline with 2 delay slots */

  {
    gm_u32_t c;

    a = from[0];
    b = from[1];
    c = from[2];
    to[0] = a;
    to[1] = b;
    to[2] = c;
    a = from[3];
    b = from[4];
    c = from[5];
    to[3] = a;
    to[4] = b;
    to[5] = c;
  }
#else
  /* optimized for LANai{3,4,5,6} pipeline with 1 delay slot. */

  a = from[0];
  b = from[1];
  to[0] = a;
  to[1] = b;
  a = from[2];
  b = from[3];
  to[2] = a;
  to[3] = b;
  a = from[4];
  b = from[5];
  to[4] = a;
  to[5] = b;
#endif
}

#endif /* _gm_routes_h_ */

