/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_list.h"

/* Print a summary of what send records are queued for the connection */

inline static
void
dump_send_records (gm_send_record_t *sr)
{
  if (!GM_DEBUG)		/* prevent silent execution of loop */
    return;

  GM_ASSERT_NO_CYCLE (gm_send_record_t, sr, next);
  
  while (sr)
  {
    _GM_PRINT (1, ("  %p sexno=0x%08x%s\n", sr, sr->sexno.whole,
		   sr->requires_explicit_ack ? " (explicit)" : ""));
    sr = sr->next;
  }
}

inline static
void
_free_send_record (gm_send_record_t *sr)
{
  gm_send_record_t *free_send_records;

  GM_PRINT (GM_DEBUG_GET, ("freeing send record %p\n", sr));
  
  free_send_records = gm.free_send_records;
  gm.free_send_records = sr;
  sr->requires_explicit_ack = 0;
  sr->next = free_send_records;
}

#define free_send_record(sr) do {					\
  GM_ASSERT_NOT_IN_LIST (gm_send_record_t,				\
			 gm.free_send_records, sr, next);		\
  _free_send_record (sr);						\
} while (0)

/****************************************************************
 * Send record recording
 *
 * Record send records corresponding to explicity acks.  This
 * allows the ack processing (in gm_control_packet.c) to communicate
 * these send records to the RDMA state machine (in gm_rdma.c).
 ****************************************************************/

inline static
void
record_required_explicit_ack (gm_send_record_t *sr)
{
  gm_send_record_t *last;

  last = gm.last_required_explicit_ack;
  gm.last_required_explicit_ack = sr;
  sr->next = 0;
  last->next = sr;
}

inline static
gm_send_record_t *
first_recorded_required_explicit_ack ()
{
  return gm.first_required_explicit_ack;
}

inline static
void
remove_first_recorded_required_explicit_ack (gm_send_record_t *sr,
					     gm_send_record_t *sr__next)
{
  gm_assert (sr);
  gm_assert (sr == gm.first_required_explicit_ack);
  gm_assert (sr__next == sr->next);
  gm.first_required_explicit_ack = sr__next;
  if (!sr__next)
    {
      gm.last_required_explicit_ack
	= (gm_send_record_t *) ((char *) &gm.first_required_explicit_ack
				- GM_OFFSETOF (gm_send_record_t, next));
    }
}

