/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_send_tokens_h_
#define _gm_send_tokens_h_

#include "gm_call_trace.h"
#include "gm_debug_send_tokens.h"
#include "gm_dispatch.h"
#include "gm_enable_debug_counters.h"
#include "gm_recv_queue.h"
#include "gm_send_tokens.h"
#include "gm_subport_cnt.h"

static inline void
update_host_sent_queue (gm_port_protected_lanai_side_t * p)
{
  unsigned int cnt;
  gm_dp_t ear;
  gm_lp_t lar;
  struct _gm_sent_token_report *p__sent_slot;

  GM_PRINT (GM_DEBUG_SEND_TOKENS,
	    ("updating sent queue for port %d\n", p->id));

  p__sent_slot = p->sent_slot;
  ;
  ;
  gm_assert ((char *) p__sent_slot + GM_NUM_SEND_QUEUE_SLOTS + 2
	     >= (char *) &p->sent + 1);

  /* null terminate the list */

  gm_assert (p->sent_slot);
  p__sent_slot->token = 0;

  /* determine event to DMA */

  cnt =
    GM_RDMA_ROUNDUP (u32, (char *) (&p->sent + 1) - (char *) p__sent_slot);
  lar = (char *) (&p->sent + 1) - cnt;

  if (GM_DEBUG_SEND_TOKENS)
    {
      if (cnt > 1)
	{
	  LOG_DISPATCH (0, "passing multiple send tokens to host");
	}
      else
	{
	  LOG_DISPATCH (0, "passing 1 send tokens to host");
	}
    }

  GM_PRINT (GM_DEBUG_SEND_TOKENS, ("event looks like:"));
  if (GM_DEBUG_SEND_TOKENS)
    {
      gm_hex_dump (lar, cnt);
    }

  /* determine where to DMA it. */

  ear = (RECV_QUEUE_SLOT_DMA_ADDR (p) + sizeof (gm_recv_queue_slot_t) - cnt);

  /* DMA the GM_SENT_EVENT to the host. */

  synchronous_GRANULAR_RDMA (lar, ear, cnt);

  RECV_QUEUE_ADVANCE (p);

  /* If needed inform the host of the completed sends. */
  wake_port_if_needed (p);

  /* Rewind the sent list. */
  p->sent_slot = 0;
}

/* Transfer the sent message lists that have been built for each port
   to the host, generating interrupts if needed.  This function is
   called periodically by the timer handler.  The sent message
   notification is batched because this information is not latency
   sensitive, batching utilizes the I/O bus more efficiently, and
   having the queue in host memory allows the host to more efficiently
   poll for sent messages.  */

static inline void
update_host_sent_queues (void)
{
  if (gm.first_port_with_sent_packets)
    {
      gm_port_protected_lanai_side_t *p;

      p = gm.first_port_with_sent_packets;
      gm.first_port_with_sent_packets = 0;
      do
	{
	  update_host_sent_queue (p);
	  /* Remove port from list of ports with sent messages. */
	  p = p->next_with_sent_packets;
	}
      while (p);
    }
}

static inline struct gm_send_queue_slot *
FAKE_SEND_QUEUE_SLOT (void)
{
  gm_assert (gm.volatile_zero == 0);
  gm_assert (GM_NO_SEND_EVENT == 0);
  return ((struct gm_send_queue_slot *) (&gm.volatile_zero + 1) - 1);
}

static inline int
send_token_in_free_list (gm_send_token_t * st,
			 gm_port_protected_lanai_side_t * port)
{
  gm_send_token_t *to_free;
  int cnt=0;

  for (to_free = port->first_free_send_token;
       to_free;
       to_free = to_free->common.next)
    {
      if (to_free == st)
	return 1;
      gm_assert (cnt++ < GM_NUM_SEND_QUEUE_SLOTS);
    }
  return 0;
}

/* Pass the sent token back to the host */

static inline void
pass_sent_token_to_port (unsigned int token,
			 gm_port_protected_lanai_side_t * port,
			 gm_boolean_t force_update, gm_status_t status)
{
  struct _gm_sent_token_report *port__sent_slot;
  unsigned int port__wake_host;
  gm_port_protected_lanai_side_t *gm_first_port_with_sent_packets;
  gm_u32_t isr;

  GM_INCR_DEBUG_CNT (gm.sent_tokens_queued_for_host_cnt);

start_over:

  GM_PRINT (GM_DEBUG_SEND_TOKENS,
	    ("passing token for slot 0x%x to port\n", token - 1));
  gm_assert (token);
  gm_assert (token <= GM_NUM_SEND_QUEUE_SLOTS);
  gm_assert (token <= 255);
  /* pre */ port__sent_slot = port->sent_slot;
  /* pre */ gm_first_port_with_sent_packets = gm.first_port_with_sent_packets;
  /* pre */ port__wake_host = port->wake_host;

  /* If needed, add this port to the list of ports with sent messages
     to be passed to the host. */

  if (port__sent_slot == 0)
    {
      GM_PRINT (GM_DEBUG_SEND_TOKENS, ("starting new sent report list\n"));

      port__sent_slot
	= &port->sent.report[GM_NUM_ELEM (port->sent.report) - 1];
      port->next_with_sent_packets = gm_first_port_with_sent_packets;
      gm.first_port_with_sent_packets = port;
    }

  /* If there is no room for the report, then generate a sent tokens
     event to reset the list and try again. */
  else if (port__sent_slot == &port->sent.report[0])
    {
      GM_PRINT (GM_DEBUG_SEND_TOKENS, ("flushing full sent report list\n"));

      update_host_sent_queues ();
      goto start_over;
    }


  /* pre */ isr = get_ISR ();
  port__sent_slot->token = token;
  port__sent_slot->status = status;
  port__sent_slot--;
  port->sent_slot = port__sent_slot;

  /* If the DMA interface is idle or a wake has been requested for
     this port, immediately pass the sent messages to the host and
     wake the port. */

  if (force_update || DMA_done () || port__wake_host)
    update_host_sent_queues ();
}

static inline void
free_send_token (gm_send_token_t * st, gm_port_protected_lanai_side_t * p)
{
  gm_send_token_t *last, *first;

  /* Free token, highly optimized for L7 pipelining */

  gm_assert (!send_token_in_free_list (st, p));

  first = p->first_free_send_token;
  last = p->last_free_send_token;
  p->last_free_send_token = st;
  st->common.next = 0;
  if (!first)
    {
      gm_port_unprotected_lanai_side_t *PORT;

      /* pre */ PORT = p->PORT;
      first = st;
      last = (gm_send_token_t *) ((char *) &p->first_free_send_token
				  - GM_OFFSETOF (gm_send_token_t,
						 common.next));
      p->send_token_queue_slot
	= &PORT->send_token_queue[st - &p->_send_tokens[0]];
      GM_PRINT (GM_DEBUG_SEND_TOKENS,
		("looking for next send in slot 0x%x\n",
		 (int) (st - &p->_send_tokens[0])));
    }
  last->common.next = st;
}

/* Queue the sent message to be passed to the host eventually via a
   bulk DMA. */

static inline void
pass_sent_token_to_port_and_free (gm_send_token_t * st,
				  gm_port_protected_lanai_side_t * port,
				  int force_sent_queue_update,
				  gm_status_t status)
{
  if (GM_ENABLE_DEBUG_COUNTERS)
    {
      if (st->common.type == GM_ST_RELIABLE)
	{
	  GM_INCR_DEBUG_CNT (gm.completed_reliable_send_cnt);
	}
      else if (st->common.type == GM_ST_DATAGRAM)
	{
	  GM_INCR_DEBUG_CNT (gm.completed_datagram_send_cnt);
	}
    }

  free_send_token (st, port);
  pass_sent_token_to_port (st - &port->_send_tokens[0] + 1, port,
			   force_sent_queue_update, status);
}

static inline void
rewind_send_tokens (gm_send_record_t * sent)
{
  /* restore send tokens to their earlier state */
  do
    {
      gm_send_token_t *st;
      gm_send_record_t *sent__next;
      gm_u32_t before_len, send_len;
      gm_up_t before_ptr;

      st = sent->send_token;
      before_len = sent->before_len;
      before_ptr = sent->before_ptr;
      send_len = st->ackable.send_len;
      sent__next = sent->next;
      ;

      /* Only rewind the send token if it is not already rewound,
         because an earlier send record may have rewound it already.
         Rewinding to the position in the later send record would be
         incorrect. */
      if (send_len < before_len)
	{
	  st->ackable.send_len = before_len;
	  st->ackable.send_ptr = before_ptr;
	}
      sent = sent__next;
    }
  while (sent);
}

/****************************************************************
 * SDMA support functions
 ****************************************************************/

/* Commit to using the first send token by removing it from the free list */

static inline void
gm_send_token_commit (gm_port_protected_lanai_side_t * port,
		      gm_send_token_t * st)
{
  gm_send_token_t *next;
  gm_port_unprotected_lanai_side_t *PORT;

  /* remove the send token from the free list */

  gm_assert (st);
  gm_assert (port->first_free_send_token == st);
  next = st->common.next;
  PORT = port->PORT;
  st->common.next = 0;
  port->first_free_send_token = next;
  if (next)
    {
      /* record the next send token slot to poll.  ??? should we do
         this is pass_sent_token_to_port_and_free? */
      port->send_token_queue_slot
	= &PORT->send_token_queue[next - &port->_send_tokens[0]];
      GM_PRINT (GM_DEBUG_SEND_TOKENS,
		("looking for next send in slot 0x%x.\n",
		 (int) (next - &port->_send_tokens[0])));
      LOG_DISPATCH (0, "gm_send_token_commit:next send slot set");
    }
  else
    {
      gm_putstring ("gm_recycle_first_send_token: faking\n");

      /* make the MCP poll a fake host send token location where it is
         guaranteed not to find a send. */
      gm_assert (GM_NO_SEND_EVENT == 0);
      port->send_token_queue_slot = FAKE_SEND_QUEUE_SLOT ();
      GM_PRINT (GM_DEBUG_SEND_TOKENS, ("out of send tokens\n"));
      LOG_DISPATCH (0, "out of send tokens");
    }
}

/* Commit to using the first send token by removing it from the free
   list */

static inline void
gm_recycle_first_send_token (gm_port_protected_lanai_side_t * port,
			     gm_send_token_t * st)
{
  gm_send_token_commit (port, st);

  /* recycle it */

  pass_sent_token_to_port_and_free (st, port, 1, GM_SUCCESS);
}

/****************************************************************
 * Send queue
 ****************************************************************/

/****************
 * Send queueing
 ****************/

static inline void
append_send_token_to_send_queue (gm_send_token_t * st, unsigned sp_id,
				 unsigned target_node_id)
{
  gm_connection_t *c;
  gm_subport_t *sp;
  gm_u32_t c__active_subport_bitmask;
  gm_subport_t *gm_free_subports;
  gm_u32_t port_id;

  gm_assert (st);
  gm_assert (target_node_id <= gm.max_node_id);
  check_subport_cnt ();
  GM_INCR_DEBUG_CNT (gm.sends_in_send_queue_cnt);

  c = &gm_connection[target_node_id];
  c__active_subport_bitmask = c->active_subport_bitmask;
  /* speculative prefetch */ gm_free_subports = gm.free_subports;
  /* speculative prefetch */ port_id = GM_SUBPORT_PORT (sp_id);
  if (c__active_subport_bitmask)
    {
      if (GM_DEBUG_SEND_QUEUE)
	gm_assert (gm_debug_send_queue_cnt);

      /* The connection is active and enqueued */
      if (c->active_subport_bitmask & (1 << sp_id))
	{
	  /* The subport is active, allocated, */
	  /* initialized, and queued. */

	  gm_send_token_t *sp__last_send_token;

	  gm_assert (c->first_active_send_port);
	  sp = c->first_active_send_port;
	  ;
	  ;
	  if (sp)		/* may be 0 for FT HACK. */
	    {
	      while (sp->id != sp_id)
		sp = sp->next;
	    }
	  gm_assert (sp->last_send_token);
	  sp__last_send_token = sp->last_send_token;

	  /* Attach the send token to the subport */
	  st->common.subport = sp;

	  /* Append send token to the send list */
	  gm_assert (sp->last_send_token != st);
	  gm_assert (sp->last_send_token->common.next == 0);
	  sp->last_send_token = st;
	  sp__last_send_token->common.next = st;
	  gm_assert (st->common.next == 0);
	}
      else
	{
	  /* The subport is NOT active, */
	  /* allocated, initialized, or queued. */

	  gm_subport_t *forward, *backward;
	  gm_subport_t *sp__next;
	  gm_u32_t gm_port___active_subport_cnt;
	  gm_s32_t rtc;

	  /* Allocate a subport */
	  gm_assert (gm_free_subports);
	  gm_assert (gm_free_subports == gm.free_subports);
	  sp = gm_free_subports;
	  gm_assert (!sp->disabled);
	  /* pre */ sp__next = sp->next;
	  gm_port___active_subport_cnt = gm_port[port_id].active_subport_cnt;


	  gm.free_subports = gm_free_subports = sp__next;

	  /* BAD: could be eliminated... somehow */
	  gm_port[port_id].active_subport_cnt
	    = gm_port___active_subport_cnt + 1;

	  /* Attach the send token to the subport */
	  st->common.subport = sp;

	  /* Initialize the subport */
	  sp->id = sp_id;
	  sp->connection = c;
	  gm_assert (sp->first_send_token == 0);
	  rtc = RTC;
	  sp->first_send_token = sp->last_send_token = st;
	  /* pre */ c__active_subport_bitmask = c->active_subport_bitmask;
	  forward = c->first_active_send_port;
	  gm_assert (st->common.next == 0);
	  gm_assert (sp->disabled == 0);
	  sp->delay_until = rtc;
	  sp->progress_time = rtc;
	  backward = forward->prev;
	  c->active_subport_bitmask =
	    c__active_subport_bitmask | (1 << sp_id);

	  gm_assert (c->first_active_send_port);
	  gm_assert (c->first_active_send_port->next);
	  gm_assert (c->first_active_send_port->prev);

	  sp->next = forward;
	  sp->prev = backward;
	  forward->prev = backward->next = sp;
	  /* Connection already in active list */

	  incr_subport_cnt ();
	}
    }
  else
    {
      /* The connection is not active or queued */

      gm_subport_t *sp__next;
      gm_u32_t gm_port___active_subport_cnt;
      gm_s32_t rtc;
      gm_connection_t *__gm_first_active_connection;

      /* Allocate a subport */
      gm_assert (gm_free_subports);
      gm_assert (gm_free_subports == gm.free_subports);
      sp = gm_free_subports;
      gm_assert (!sp->disabled);
      /* pre */ sp__next = sp->next;
      /* pre */ gm_port___active_subport_cnt
	= gm_port[port_id].active_subport_cnt;
      gm.free_subports = gm_free_subports = sp__next;

      gm_port[port_id].active_subport_cnt = gm_port___active_subport_cnt + 1;

      /* Attach the send token to the subport */
      st->common.subport = sp;

      /* Initialize the subport */
      sp->id = sp_id;
      sp->connection = c;
      gm_assert (sp->first_send_token == 0);
      rtc = RTC;
      sp->first_send_token = sp->last_send_token = st;
      gm_assert (st->common.next == 0);
      gm_assert (sp->disabled == 0);
      sp->delay_until = rtc;
      sp->progress_time = rtc;
      c->known_alive_time = rtc;

      /* Insert the subport in the active list for the connection */
      /* pre */ __gm_first_active_connection = gm.first_active_connection;
      c->active_subport_bitmask = 1 << sp_id;
      sp->next = sp->prev = sp;
      c->first_active_send_port = sp;

      /* Insert connection in active connection list if needed. */
      if (__gm_first_active_connection)
	{
	  gm_connection_t *forw, *back;

	  if (GM_DEBUG_SEND_QUEUE)
	    gm_assert (gm_debug_send_queue_cnt);

	  forw = __gm_first_active_connection;
	  back = forw->prev_active;
	  c->next_active = forw;
	  forw->prev_active = back->next_active = c;
	  c->prev_active = back;
	}
      else
	{
	  if (GM_DEBUG_SEND_QUEUE)
	    gm_assert (!gm_debug_send_queue_cnt);

	  gm.first_active_connection = c;
	  c->next_active = c->prev_active = c;
	}

      incr_subport_cnt ();
    }

  gm_assert (st);
  gm_assert (st->common.next == 0);
  gm_assert (sp);
  gm_assert (c);
  gm_assert (st->common.subport == sp);
  gm_assert (sp->connection == c);
  gm_assert (sp->id == sp_id);
  gm_assert (sp->first_send_token);
  gm_assert (sp->last_send_token == st);
  gm_assert (c->active_subport_bitmask & (1 << sp_id));
  check_subport_cnt ();
}

static inline void
remove_first_send_token_from_send_queue (gm_send_token_t * st, unsigned sp_id)
{
  gm_subport_t *sp;
  gm_subport_t *sp__next;
  gm_connection_t *c;
  gm_send_token_t *sp__first_send_token;
  gm_subport_t *gm_free_subports;
  gm_u32_t gm_port___active_subport_cnt;
  gm_send_token_t *st__common_next;

  check_subport_cnt ();
  GM_DECR_DEBUG_CNT (gm.sends_in_send_queue_cnt);

  gm_assert (st);
  /* Remove send token from send list */
  sp = st->common.subport;
  /* pre */ st__common_next = st->common.next;
  /* pre */ gm_free_subports = gm.free_subports;
  /* pre */ c = sp->connection;

  gm_assert (sp);
  gm_assert (sp->first_send_token);
  gm_assert (sp->first_send_token == st);
  gm_assert (sp->id == sp_id);
  gm_assert (sp->connection);
  gm_assert (sp->connection->active_subport_bitmask & (1 << sp_id));

  sp->first_send_token = sp__first_send_token = st__common_next;
  /* pre */ sp__next = sp->next;
  if (!sp__first_send_token)
    {
      gm_connection_t *c__next_active;
      gm_u32_t sp__id;

      /* Remove the subport from the send queue. */

      gm_assert (c);
      gm_assert (c == sp->connection);

      /* prefetch */ c__next_active = c->next_active;
      /* prefetch */ sp__id = sp->id;

      gm_assert (sp__next);
      if (sp__next == sp)
	{
	  gm_connection_t *gm_first_active_connection;

	  /* Remove the last subport for the connection. */
	  /* Mark connection as inactive */
	  gm_assert (c->active_subport_bitmask == 1U << sp_id);
	  /* pre */ gm_first_active_connection = gm.first_active_connection;
	  c->active_subport_bitmask = 0;

	  /* Remove connection from send list, marking the send list
	     as empty if it is empty. */
	  if (c__next_active == c)
	    {
	      gm.first_active_connection = 0;
	      NOTICE_NO (SDMA_PENDING);
	    }
	  else
	    {
	      gm_connection_t *forward, *backward;
	      backward = c->prev_active;
	      gm_assert (gm_first_active_connection);
	      if (gm_first_active_connection == c)
		gm.first_active_connection = c__next_active;

	      forward = c__next_active;
	      gm_assert (forward);
	      gm_assert (backward);
	      forward->prev_active = backward;
	      backward->next_active = forward;
	    }
	}
      else
	{
	  /* Remove a subport for the connection, but it's not the last one. */
	  gm_subport_t *forward, *backward;
	  gm_u32_t c__active_subport_bitmask;

	  backward = sp->prev;
	  c__active_subport_bitmask = c->active_subport_bitmask;
	  gm_assert (c->first_active_send_port);
	  forward = sp__next;
	  if (c->first_active_send_port == sp)
	    c->first_active_send_port = sp__next;

	  gm_assert (forward);
	  gm_assert (backward);
	  forward->prev = backward;
	  backward->next = forward;

	  gm_assert (c->active_subport_bitmask & 1 << sp_id);
	  c->active_subport_bitmask =
	    c__active_subport_bitmask ^ (1 << sp_id);
	}
      /* Free the subport. */

      /* pre */ gm_port___active_subport_cnt
	= gm_port[GM_SUBPORT_PORT (sp__id)].active_subport_cnt;

      sp->next = gm_free_subports;
      gm.free_subports = sp;

      /* BAD: could be eliminated... somehow */
      gm_port[GM_SUBPORT_PORT (sp_id)].active_subport_cnt
	= gm_port___active_subport_cnt - 1;

      decr_subport_cnt ();
    }

  check_subport_cnt ();

  LOG_DISPATCH (0, "removed first send token from send queue");
}

/* Generic remove_send_token_from_send_queue() routine.  This is used
   to remove datagram send tokens from the queue, since the
   datagram send token we want to remove may not be at the head
   of the send queue; reliable send tokens may precede it. */

static inline void
remove_send_token_from_send_queue (gm_send_token_t *st, unsigned sp_id)
{
  gm_send_token_t *sp__first_send_token;
  gm_send_token_t *st__common_next;
  gm_subport_t *sp;

  sp = st->common.subport;
  st__common_next = st->common.next;
  ;
  sp__first_send_token = sp->first_send_token;
  ;
  ;
  gm_assert (sp__first_send_token);
  if (sp__first_send_token == st)
    {
      /* st is the first in the list */

      remove_first_send_token_from_send_queue (st, sp_id);
    }
  else
    {
      gm_send_token_t *scan_token;

      /* find the token containing a COMMON.NEXT pointer to ST */

      for (scan_token = sp__first_send_token;
	   scan_token->common.next != st;
	   scan_token = scan_token->common.next)
	{
	  gm_assert (scan_token->common.next);
	}

      /* Splice ST out of the linked list of send tokens. */

      scan_token->common.next = st__common_next;

      /* If ST is the last send token for the subport, update
         SP->LAST_SEND_TOKEN. */

      if (st__common_next == 0)
	{
	  sp->last_send_token = scan_token;
	}

      LOG_DISPATCH (0, "removed send token from send queue (not first)");
      GM_DECR_DEBUG_CNT (gm.sends_in_send_queue_cnt);
    }
}

/* rotate the send queue given the head suport and connection */

static inline void
rotate_send_queue (gm_subport_t * sp, gm_connection_t * c)
{
  gm_subport_t *sp__next;
  gm_connection_t *c__next_active;

  gm_assert (c);
  gm_assert (c == gm.first_active_connection);
  gm_assert (sp);
  gm_assert (c->first_active_send_port == sp);

  sp__next = sp->next;
  c__next_active = c->next_active;
  ;
  gm_assert (sp__next);
  gm_assert (c__next_active);
  c->first_active_send_port = sp__next;
  gm.first_active_connection = c__next_active;
}

#endif /* _gm_send_tokens_h_ */
