/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gmcp_h_
#define _gmcp_h_

#include "gm_config.h"
#include "gm_types.h"
#include "gm_debug_recv_tokens.h"
#include "gm_bootstrap.h"

/****************
 * Compact switches
 ****************/

#define GM_STR(str) (str)

/***********************************************************************
 * Finding GM LANai globals
 ***********************************************************************/

/* The "gm" structure includes most of the LANai globals, except those
   that are mapped into host memory.

   "GM" is an array of user-writable page-sized data structures
   associated with each port and mapped into host memory.


         ...
   |                       |
   |-----------------------|
   | GM_PORT(0)            |
   |-----------------------|
         ...
   |-----------------------|
   | GM_PORT(GM_NUM_PORTS) |
   |-----------------------|
   |                       |
   | Page hash piece       |
   |                       |
   +-----------------------+ (end of memory) */

#define GM_PORT (gm._PORT)
#define GM_PAGE_HASH_PIECE ((gm_dp_t *)&GM_PORT[GM_NUM_PORTS])

#define GM_STACK_WORDS 8*1024
extern gm_u32_t gm_stack[GM_STACK_WORDS];
extern void *gm_top_of_stack;

/****************************************************************
 * 
 ****************************************************************/

/* This should be a power of 2 that is approximately the number of
   bytes the network can transfer in 1 microsecond. */

#define GM_BYTES_PER_USEC 128

#if GM_DEBUG
/* If debugging is on, the MCP runs so slowly that the timer period
   needs to be extended to prevent the timer from consuming all cycles.
   Also, be abbreviate some of the longer timeouts so errors will be
   reported more quickly. */
#define GM_TIMER_PERIOD 2000
#define GM_LOW_FREQ_TIMER_PERIOD (4000)
#define GM_SEND_TIMEOUT (2*1000*1000)
#define GM_FATAL_SEND_TIMEOUT (10*1000*1000*2)
#else /* !GM_DEBUG */

/* These are the normal timer values */
#define GM_TIMER_PERIOD 200
#define GM_LOW_FREQ_TIMER_PERIOD (2000)
#define GM_SEND_TIMEOUT 4096
#define GM_FATAL_SEND_TIMEOUT (30*1000*1000*2)

#endif /* !GM_DEBUG */

/* This must be substantially smaller than GM_FATAL_SEND_TIMEOUT */
#define GM_CONNECTION_TIMEOUT (GM_FATAL_SEND_TIMEOUT/2)

/****************************************************************
 * 
 ****************************************************************/

/****************
 * Sequence number support
 ****************/

/* Compare 32-bit sequence numbers. Return 0 if equal, negative if
   a<b, and positive if a>b, but compensating for wraparound. */

#define SEQ_CMP32(a, b) ((gm_s32_t) (a) - (gm_s32_t) (b))

/* Compare 16-bit sequence numbers. Return 0 if equal, negative if
   a<b, and positive if a>b, but compensating for wraparound */

#define SEQ_CMP16(a, b) (((gm_s32_t) (a) - (gm_s32_t) (b)) << 16)

/****************************************************************
 * Functions exported by gmcp.c
 ****************************************************************/

void gm_reverse_bcopy (void *from, void *to, unsigned long len);
/* void _gm_bcopy_down (const void *from, void *to, gm_size_t len); */
/* void _gm_bcopy_up (const void *from, void *to, gm_size_t len); */
void _gm_overlapping_bcopy (void *from, void *to, gm_size_t len);
void gm_putstring (char *my_str);


#endif /* _gmcp_h_ */

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
